#!/usr/bin/env python

import json

from os.path import join
from pathlib import Path
from typing import Any, Dict, List, Optional, TypedDict, Union

from datetime import date

import arrow
import click
import requests
import string

from canvas_workflow_kit import settings
from canvas_workflow_kit.api import fetch_patient
from canvas_workflow_kit.patient import Patient
from canvas_workflow_kit.protocol import CONTEXT_REPORT, CONTEXT_GUIDANCE
from canvas_workflow_kit.timeframe import Timeframe
from canvas_workflow_kit.internal.string import snakecase
from canvas_workflow_kit.utils import (
    parse_class_from_python_source, load_local_patient, load_patient
)

mocks_path = 'TODO'


# Utility methods -----------
#
def green(string: str) -> str:
    return click.style(string, fg='green')


def blue(string: str) -> str:
    return click.style(string, fg='blue', bold=True)


def cyan(string: str) -> str:
    return click.style(string, fg='cyan', bold=True)


def yellow(string: str) -> str:
    return click.style(string, fg='yellow', bold=True)


def red(string: str) -> str:
    return click.style(string, fg='red')


# CLI Commands -----------------------------------

@click.group()
@click.pass_context
@click.option('--use-config', required=False)
def cli(ctx, use_config='canvas_cli'):

    ctx.ensure_object(dict)
    ctx.obj['config_section'] = use_config and use_config or settings.DEFAULT_CONFIG_SECTION

    if ctx.invoked_subcommand != 'create-default-settings':
        try:
            ctx.obj['settings'] = settings.read_config(ctx.obj['config_section'])
        except (FileNotFoundError, ValueError) as e:
            raise click.ClickException(e)


@cli.command()
def create_default_settings():
    """
    Create a default settings file with placeholder text in `~/.canvas/config.ini`.
    File will only be created if it does not yet exist.
    """

    settings_path = settings.CONFIG_PATH

    if settings_path.is_file():
        raise click.ClickException(f'Settings file already exists at {settings_path}')

    settings_path.parent.mkdir(parents=True, exist_ok=True)

    click.echo(f'Writing default settings to "{settings_path}"...')

    settings_path.write_text('''[canvas_cli]
url =
api-key =
''')


@cli.command()
@click.argument('output_path')
def create(output_path: Path):
    """
    Create a new item
    """
    output_path = Path(output_path)
    if output_path.is_file():
        raise click.ClickException(f'File already exists at {output_path}')

    template_path = Path(__file__).parent / 'builtin_cqms/stub_template_user.py.txt'
    template = template_path.open('r').read()

    content = template.format(**{
        'value_sets': '*',
        'year': date.today().year,
        'class_name': "MyClinicalQualityMeasure",
        'title': "My Clinical Quality Measure",
        'description': 'Description',
        'information_link': 'https://link_to_protocol_information',
        'types': ['CQM'],
        'authors': '"Canvas Example Medical Association (CEMA)"',
        'references': '"Gunderson, Beau;  Excellence in Software Templates. 2012 Oct 20;161(7):422-34."',
        'funding_source': '',
        'denominator': 'pass',
        'numerator': 'pass'
    })
    with output_path.open('w') as output_handle:
        output_handle.write(content)

    click.echo(green(f'Successfully wrote template to {output_path.absolute()}'))


@cli.command()
@click.argument('patient-keys', nargs=-1)
@click.option('-d', '--destination', required=False, type=click.Path(),
              help="Specify a directory to store the output.")
@click.pass_context
def fixture_from_patient(ctx, patient_keys: List[str], destination: Path):
    """
    Export a fixture for the list of provided patient keys.
    Fixtures will be created with files representing the various categories.
    Eg: billingLineItems.json, conditions.json, referralReports.json ...
    """

    for patient_key in patient_keys:
        _process_fixture_from_patient(patient_key, destination)


@click.pass_context
def _process_fixture_from_patient(ctx, patient_key, output_directory):
    click.echo(f'Getting fixture from patient "{patient_key}"...')

    response_json = fetch_patient(patient_key, ctx.obj['config_section'])

    if not output_directory:
        output_directory = Path(".")

    patient = response_json['patient']
    patient_summary = f"{patient['firstName']} {patient['lastName']} {patient['birthDate'][0:4]} ({patient['sexAtBirth']})"

    output_directory = Path(output_directory) / patient_summary

    output_directory.mkdir(parents=True, exist_ok=True)
    for key, values in response_json.items():
        (output_directory / f'{snakecase(key)}.json').write_text(json.dumps(values))

    click.echo(green(f'Successfully wrote patient fixture to {output_directory.absolute()}'))


@cli.command()
@click.argument('module-path')
@click.argument('fixture-folder')
@click.option('--date')
@click.option('--start-date')
@click.option('--end-date')
@click.option('--report-mode', default=False, is_flag=True, help="Run the class in report mode.")
@click.option('--full', flag_value='full', default=False,
              help="Show the full output returned to the application.")
@click.pass_context
def test_fixture(
    ctx,
    module_path: str, fixture_folder: str,
    date: str = None, start_date: str = None, end_date: str = None,
    report_mode: bool = False,
    full: bool = False
):
    """
    Test a python file with a ClinicalQualityMeasure against a fixture folder.
    """

    module_path = Path(module_path)
    try:
        Class = parse_class_from_python_source(module_path.open('r').read())
    except SyntaxError as e:
        if not e.text:
            e.text = ''
        raise SyntaxError(
            f'Could not parse python file.\n  File "{module_path.absolute()}", line {e.lineno}\n    {e.text.strip()}\nSyntaxError: {e.msg}')

    if len(fixture_folder) == 32 and set(fixture_folder).issubset(string.hexdigits):
        patient = load_patient(fixture_folder, ctx.obj.get('config_section'))
        test_patient(patient, Class, date, start_date, end_date, report_mode, full)
        return

    path = Path(fixture_folder)

    subdirectories = [x for x in path.iterdir() if x.is_dir()]

    if subdirectories:
        for i, fixture_folder in enumerate(subdirectories):
            if i >= 1:
                click.echo('-' * 80)

            # 2. load JSON folder of fixture data
            fixture_title = f"Fixture: {yellow(fixture_folder.name)}"
            click.echo(fixture_title)
            patient = load_local_patient(Path(fixture_folder))

            test_patient(patient, Class, date, start_date, end_date, report_mode, full)

    else:
        fixture_title = f"Fixture: {yellow(fixture_folder.name)}"
        click.echo(fixture_title)
        patient = load_local_patient(path)
        test_patient(patient, Class, date, start_date, end_date, report_mode, full)


def test_patient(
    patient: Patient, Class: type,
    date: str = None, start_date: str = None, end_date: str = None,
    report_mode: bool = False,
    full: bool = False
):

    if date:
        date = arrow.get(date)
    else:
        date = arrow.now()

    if start_date:
        start_date = arrow.get(start_date)
    else:
        start_date = arrow.now().shift(years=-1)

    if end_date:
        end_date = arrow.get(end_date)
    else:
        end_date = arrow.now()

    timeframe = Timeframe(start=start_date, end=end_date)

    # 3. instantiate module
    protocol = Class(patient=patient, date=date, timeframe=timeframe)
    if report_mode:
        protocol.context = CONTEXT_REPORT
    else:
        protocol.context = CONTEXT_GUIDANCE

    results = protocol.compute_results()

    # 4. return results
    recommendations = getattr(results, 'recommendations', None)
    if not recommendations:
        click.echo(red("No recommendations"))

    elif full:
        recommendations = [vars(r) for r in recommendations]
        click.echo(json.dumps(recommendations, indent=2))
    else:
        for r in recommendations:
            click.echo(cyan(r.__class__.__name__))
            for attr in ('title', 'key', 'button'):
                if hasattr(r, attr):
                    click.echo(f' {blue(attr.capitalize())}: {getattr(r, attr)}')


@cli.command()
@click.argument('filename', type=click.Path(exists=True))
@click.option('--force', default=False, is_flag=True, help="Allow the version to be overridden.")
@click.pass_context
def upload(ctx, filename: Path, force=False):
    """
    Upload a ClinicalQualityMeasure to the server
    """

    if not filename.endswith('.py'):
        raise click.ClickException(f'Only python files with a .py extension can be uploaded.')

    filename_path = Path(filename)

    click.echo(f'Uploading {filename_path.name}...')

    with filename_path.open() as f:
        contents = f.read()

    files = {'file': filename_path.open('rb')}

    # Send additional data with GET.  Using POST messes up the file upload.
    force_str = ''
    if force:
        force_str = '?force=1'

    response = requests.post(join(
        ctx.obj["settings"]["url"], f'api/PatientProtocol/upload/'
    ) + force_str, files=files, headers={
        'Authorization': ctx.obj['settings']['api_key'],
        'Content-Length': str(len(contents)),
        'Content-Disposition': f'attachment; filename="{filename_path.name}"'
    })

    if response.status_code == 201 and response.json().get('status') == 'success':
        version = response.json().get('data', {}).get('version')
        version_str = ''
        if version:
            version_str = f'Version {version}'
        click.echo(green(f'Upload successful.  {version_str} set to latest version.'))
    else:
        raise click.ClickException(response.text)


@cli.command()
@click.argument('module-name')
def set_active(module_name: str):
    """
    Set a protocol to active on the server
    """
    click.echo(f'Setting {module_name} as active...')
    _set_active(True, module_name)


@cli.command()
@click.argument('module-name')
def set_inactive(module_name: str):
    """
    Set a protocol to inactive on the server
    """
    click.echo(f'Setting {module_name}" as inactive...')
    _set_active(False, module_name)


@click.pass_context
def _set_active(ctx, is_active: bool, module_name: str):
    response = requests.post(join(
        ctx.obj["settings"]["url"], f'api/PatientProtocol/set_active/'
    ), data={
        'is_active': is_active and 1 or 0,
        'module_name': module_name,
    }, headers={
        'Authorization': ctx.obj['settings']['api_key'],
    })
    if response.status_code == 200:
        click.echo(green(response.json().get('data', {}).get('detail')))
    else:
        raise click.ClickException(response.text)


@cli.command()
@click.argument('module-name')
@click.pass_context
def list_versions(ctx, module_name: str):
    """
    List the available versions on the server.
    """
    response = requests.get(join(
        ctx.obj["settings"]["url"], f'api/PatientProtocol/version/{module_name}/'
    ), headers={
        'Authorization': ctx.obj['settings']['api_key'],
    })

    if response.status_code == 200:
        is_active = response.json().get('is_active')
        active_version = response.json().get('active_version')

        color_method = green if is_active else red

        click.echo("Is Active: " + color_method(is_active))
        click.echo("Active version: " + green(active_version))
        click.echo("Versions: ")

        for version in response.json().get('versions', []):
            version_number = version.get('version')
            def color_method(x): return x
            if version_number == active_version:
                color_method = green

            click.echo(color_method(f' {version_number}: {version.get("changelog")}'))

    else:
        raise click.ClickException(response.text)


@cli.command()
@click.argument('module-name')
@click.argument('version')
@click.pass_context
def set_version(ctx, module_name: str, version: str):
    """
    Set a protocol's active version on the server.
    The protocol upload may still need to be made active after changing the version.
    """
    response = requests.post(join(
        ctx.obj["settings"]["url"], f'api/PatientProtocol/version/{module_name}/'
    ), data={
        'version': version,
    }, headers={
        'Authorization': ctx.obj['settings']['api_key'],
    })
    if response.status_code == 200:
        click.echo(green(response.json().get('data', {}).get('detail')))
    else:
        raise click.ClickException(response.text)


if __name__ == '__main__':
    cli()



from ..value_set import ValueSet


class Ethnicity(ValueSet):
    """
    **Clinical Focus:** 

    **Data Element Scope:** 

    **Inclusion Criteria:** 

    **Exclusion Criteria:** 

    ** Used in:** CMS135v10, CMS90v11, CMS249v4, CMS159v10, CMS349v4, CMS69v10, CMS142v10, CMS134v10, CMS177v10, CMS165v10, CMS143v10, CMS146v10, CMS157v10, CMS124v10, CMS2v11, CMS50v10, CMS22v10, CMS68v11, CMS139v10, CMS129v11, CMS154v10, CMS161v10, CMS56v10, CMS74v11, CMS645v5, CMS138v10, CMS75v10, CMS137v10, CMS136v11, CMS128v10, CMS122v10, CMS153v10, CMS144v10, CMS66v10, CMS130v10, CMS646v2, CMS155v10, CMS127v10, CMS771v3, CMS117v10, CMS131v10, CMS149v10, CMS347v5, CMS133v10, CMS156v10, CMS147v11, CMS125v10, CMS145v10
    """

    VALUE_SET_NAME = 'Ethnicity'
    OID = '2.16.840.1.114222.4.11.837'
    DEFINITION_VERSION = '20121025'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    CDCREC = {
        '2135-2',  # Hispanic or Latino
        '2186-5',  # Not Hispanic or Latino
    }


class OncAdministrativeSex(ValueSet):
    """
    **Clinical Focus:** Gender identity restricted to only Male and Female used in administrative situations requiring a restriction to these two categories.

    **Data Element Scope:** Gender

    **Inclusion Criteria:** Male and Female only.

    **Exclusion Criteria:** Any gender identity that is not male or female.

    ** Used in:** CMS135v10, CMS90v11, CMS249v4, CMS159v10, CMS349v4, CMS69v10, CMS142v10, CMS134v10, CMS177v10, CMS165v10, CMS143v10, CMS146v10, CMS157v10, CMS124v10, CMS2v11, CMS50v10, CMS22v10, CMS68v11, CMS139v10, CMS129v11, CMS154v10, CMS161v10, CMS56v10, CMS74v11, CMS645v5, CMS138v10, CMS75v10, CMS137v10, CMS136v11, CMS128v10, CMS122v10, CMS153v10, CMS144v10, CMS66v10, CMS130v10, CMS646v2, CMS155v10, CMS127v10, CMS771v3, CMS117v10, CMS131v10, CMS149v10, CMS347v5, CMS133v10, CMS156v10, CMS147v11, CMS125v10, CMS145v10
    """

    VALUE_SET_NAME = 'ONC Administrative Sex'
    OID = '2.16.840.1.113762.1.4.1'
    DEFINITION_VERSION = '20150331'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    ADMINISTRATIVEGENDER = {
        'F',  # Female
        'M',  # Male
    }


class Payer(ValueSet):
    """
    **Clinical Focus:** Categories of types of health care payor entities as defined by the US Public Health Data Consortium SOP code system

    **Data Element Scope:** @code in CCDA r2.1 template Planned Coverage [act: identifier urn:oid:2.16.840.1.113883.10.20.22.4.129 (open)] DYNAMIC

    **Inclusion Criteria:** All codes in the code system

    **Exclusion Criteria:** none

    ** Used in:** CMS135v10, CMS90v11, CMS249v4, CMS159v10, CMS349v4, CMS69v10, CMS142v10, CMS134v10, CMS177v10, CMS165v10, CMS143v10, CMS146v10, CMS157v10, CMS124v10, CMS2v11, CMS50v10, CMS22v10, CMS68v11, CMS139v10, CMS129v11, CMS154v10, CMS161v10, CMS56v10, CMS74v11, CMS645v5, CMS138v10, CMS75v10, CMS137v10, CMS136v11, CMS128v10, CMS122v10, CMS153v10, CMS144v10, CMS66v10, CMS130v10, CMS646v2, CMS155v10, CMS127v10, CMS771v3, CMS117v10, CMS131v10, CMS149v10, CMS347v5, CMS133v10, CMS156v10, CMS147v11, CMS125v10, CMS145v10
    """

    VALUE_SET_NAME = 'Payer'
    OID = '2.16.840.1.114222.4.11.3591'
    DEFINITION_VERSION = '20180718'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    SOP = {
        '1',  # MEDICARE
        '2',  # MEDICAID
        '3',  # OTHER GOVERNMENT (Federal/State/Local) (excluding Department of Corrections)
        '4',  # DEPARTMENTS OF CORRECTIONS
        '5',  # PRIVATE HEALTH INSURANCE
        '6',  # BLUE CROSS/BLUE SHIELD
        '7',  # MANAGED CARE, UNSPECIFIED (to be used only if one can't distinguish public from private)
        '8',  # NO PAYMENT from an Organization/Agency/Program/Private Payer Listed
        '9',  # MISCELLANEOUS/OTHER
        '11',  # Medicare Managed Care (Includes Medicare Advantage Plans)
        '12',  # Medicare (Non-managed Care)
        '13',  # Medicare Hospice
        '14',  # Dual Eligibility Medicare/Medicaid Organization
        '19',  # Medicare Other
        '21',  # Medicaid (Managed Care)
        '22',  # Medicaid (Non-managed Care Plan)
        '23',  # Medicaid/SCHIP
        '25',  # Medicaid - Out of State
        '26',  # Medicaid - Long Term Care
        '29',  # Medicaid Other
        '31',  # Department of Defense
        '32',  # Department of Veterans Affairs
        '33',  # Indian Health Service or Tribe
        '34',  # HRSA Program
        '35',  # Black Lung
        '36',  # State Government
        '37',  # Local Government
        '38',  # Other Government (Federal, State, Local not specified)
        '39',  # Other Federal
        '41',  # Corrections Federal
        '42',  # Corrections State
        '43',  # Corrections Local
        '44',  # Corrections Unknown Level
        '51',  # Managed Care (Private)
        '52',  # Private Health Insurance - Indemnity
        '53',  # Managed Care (private) or private health insurance (indemnity), not otherwise specified
        '54',  # Organized Delivery System
        '55',  # Small Employer Purchasing Group
        '56',  # Specialized Stand-Alone Plan
        '59',  # Other Private Insurance
        '61',  # BC Managed Care
        '62',  # BC Insurance Indemnity
        '71',  # HMO
        '72',  # PPO
        '73',  # POS
        '79',  # Other Managed Care
        '81',  # Self-pay (Includes applicants for insurance and Medicaid applicants)
        '82',  # No Charge
        '83',  # Refusal to Pay/Bad Debt
        '84',  # Hill Burton Free Care
        '85',  # Research/Donor
        '89',  # No Payment, Other
        '91',  # Foreign National
        '92',  # Other (Non-government)
        '93',  # Disability Insurance
        '94',  # Long-term Care Insurance
        '95',  # Worker's Compensation
        '96',  # Auto Insurance (includes no fault)
        '97',  # Legal Liability / Liability Insurance
        '98',  # Other specified but not otherwise classifiable (includes Hospice - Unspecified plan)
        '99',  # No Typology Code available for payment source
        '111',  # Medicare HMO
        '112',  # Medicare PPO
        '113',  # Medicare POS
        '119',  # Medicare Managed Care Other
        '121',  # Medicare FFS
        '122',  # Medicare Drug Benefit
        '123',  # Medicare Medical Savings Account (MSA)
        '129',  # Medicare Non-managed Care Other
        '191',  # Medicare Pharmacy Benefit Manager
        '211',  # Medicaid HMO
        '212',  # Medicaid PPO
        '213',  # Medicaid PCCM (Primary Care Case Management)
        '219',  # Medicaid Managed Care Other
        '291',  # Medicaid Pharmacy Benefit Manager
        '299',  # Medicaid - Dental
        '311',  # TRICARE (CHAMPUS)
        '312',  # Military Treatment Facility
        '313',  # Dental --Stand Alone
        '321',  # Veteran care-Care provided to Veterans
        '322',  # Non-veteran care
        '331',  # Indian Health Service - Regular
        '332',  # Indian Health Service - Contract
        '333',  # Indian Health Service - Managed Care
        '334',  # Indian Tribe - Sponsored Coverage
        '341',  # Title V (MCH Block Grant)
        '342',  # Migrant Health Program
        '343',  # Ryan White Act
        '349',  # Other
        '361',  # State SCHIP program (codes for individual states)
        '362',  # Specific state programs (list/ local code)
        '369',  # State, not otherwise specified (other state)
        '371',  # Local - Managed care
        '372',  # FFS/Indemnity
        '379',  # Local, not otherwise specified (other local, county)
        '381',  # Federal, State, Local not specified managed care
        '382',  # Federal, State, Local not specified - FFS
        '389',  # Federal, State, Local not specified - Other
        '391',  # Federal Employee Health Plan - Use when known
        '511',  # Commercial Managed Care - HMO
        '512',  # Commercial Managed Care - PPO
        '513',  # Commercial Managed Care - POS
        '514',  # Exclusive Provider Organization
        '515',  # Gatekeeper PPO (GPPO)
        '516',  # Commercial Managed Care - Pharmacy Benefit Manager
        '517',  # Commercial Managed Care - Dental
        '519',  # Managed Care, Other (non HMO)
        '521',  # Commercial Indemnity
        '522',  # Self-insured (ERISA) Administrative Services Only (ASO) plan
        '523',  # Medicare supplemental policy (as second payer)
        '524',  # Indemnity Insurance - Dental
        '529',  # Private health insurance--other commercial Indemnity
        '561',  # Dental
        '562',  # Vision
        '611',  # BC Managed Care - HMO
        '612',  # BC Managed Care - PPO
        '613',  # BC Managed Care - POS
        '614',  # BC Managed Care - Dental
        '619',  # BC Managed Care - Other
        '621',  # BC Indemnity
        '622',  # BC Self-insured (ERISA) Administrative Services Only (ASO)Plan
        '623',  # BC Medicare Supplemental Plan
        '629',  # BC Indemnity - Dental
        '821',  # Charity
        '822',  # Professional Courtesy
        '823',  # Research/Clinical Trial
        '951',  # Worker's Comp HMO
        '953',  # Worker's Comp Fee-for-Service
        '954',  # Worker's Comp Other Managed Care
        '959',  # Worker's Comp, Other unspecified
        '3111',  # TRICARE Prime--HMO
        '3112',  # TRICARE Extra--PPO
        '3113',  # TRICARE Standard - Fee For Service
        '3114',  # TRICARE For Life--Medicare Supplement
        '3115',  # TRICARE Reserve Select
        '3116',  # Uniformed Services Family Health Plan (USFHP) -- HMO
        '3119',  # Department of Defense - (other)
        '3121',  # Enrolled Prime--HMO
        '3122',  # Non-enrolled Space Available
        '3123',  # TRICARE For Life (TFL)
        '3211',  # Direct Care-Care provided in VA facilities
        '3212',  # Indirect Care-Care provided outside VA facilities
        '3221',  # Civilian Health and Medical Program for the VA (CHAMPVA)
        '3222',  # Spina Bifida Health Care Program (SB)
        '3223',  # Children of Women Vietnam Veterans (CWVV)
        '3229',  # Other non-veteran care
        '3711',  # HMO
        '3712',  # PPO
        '3713',  # POS
        '3811',  # Federal, State, Local not specified - HMO
        '3812',  # Federal, State, Local not specified - PPO
        '3813',  # Federal, State, Local not specified - POS
        '3819',  # Federal, State, Local not specified - not specified managed care
        '9999',  # Unavailable / No Payer Specified / Blank
        '32121',  # Fee Basis
        '32122',  # Foreign Fee/Foreign Medical Program (FMP)
        '32123',  # Contract Nursing Home/Community Nursing Home
        '32124',  # State Veterans Home
        '32125',  # Sharing Agreements
        '32126',  # Other Federal Agency
        '32127',  # Dental Care
        '32128',  # Vision Care
    }


class Race(ValueSet):
    """
    **Clinical Focus:** 

    **Data Element Scope:** 

    **Inclusion Criteria:** 

    **Exclusion Criteria:** 

    ** Used in:** CMS135v10, CMS90v11, CMS249v4, CMS159v10, CMS349v4, CMS69v10, CMS142v10, CMS134v10, CMS177v10, CMS165v10, CMS143v10, CMS146v10, CMS157v10, CMS124v10, CMS2v11, CMS50v10, CMS22v10, CMS68v11, CMS139v10, CMS129v11, CMS154v10, CMS161v10, CMS56v10, CMS74v11, CMS645v5, CMS138v10, CMS75v10, CMS137v10, CMS136v11, CMS128v10, CMS122v10, CMS153v10, CMS144v10, CMS66v10, CMS130v10, CMS646v2, CMS155v10, CMS127v10, CMS771v3, CMS117v10, CMS131v10, CMS149v10, CMS347v5, CMS133v10, CMS156v10, CMS147v11, CMS125v10, CMS145v10
    """

    VALUE_SET_NAME = 'Race'
    OID = '2.16.840.1.114222.4.11.836'
    DEFINITION_VERSION = '20121025'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    CDCREC = {
        '1002-5',  # American Indian or Alaska Native
        '2028-9',  # Asian
        '2054-5',  # Black or African American
        '2076-8',  # Native Hawaiian or Other Pacific Islander
        '2106-3',  # White
        '2131-1',  # Other Race
    }


class Female(ValueSet):
    """
    **Clinical Focus:** Concepts that represent Female when assessing quality measures

    **Data Element Scope:** Gender

    **Inclusion Criteria:** Appropriate female gender concepts

    **Exclusion Criteria:** Concepts representing Male gender

    ** Used in:** CMS124v10, CMS153v10, CMS125v10, CMS249v4
    """

    VALUE_SET_NAME = 'Female'
    OID = '2.16.840.1.113883.3.560.100.2'
    DEFINITION_VERSION = '20160331'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    ADMINISTRATIVEGENDER = {
        'F',  # Female
    }


class White(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for the patient characteristic of white race.

    **Data Element Scope:** This value set may use a model element related to Race.

    **Inclusion Criteria:** Includes concepts that represent a patient characteristic of white race.

    **Exclusion Criteria:** Excludes concepts that represent non-white race.

    ** Used in:** CMS249v4
    """

    VALUE_SET_NAME = 'White'
    OID = '2.16.840.1.113883.3.464.1003.123.12.1007'
    DEFINITION_VERSION = '20180321'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    CDCREC = {
        '2106-3',  # White
    }




import requests

from typing import Optional

from canvas_workflow_kit import settings

from os.path import join


def fetch_patient(patient_key: str, config_name: Optional[str] = None):
    config = settings.read_config(config_name)

    response = requests.get(join(
        config['url'], f'api/PatientProtocolInput/{patient_key}'
    ), headers={'Authorization': config['api_key']})
    response.raise_for_status()
    return response.json()



# type: ignore
from typing import List, Type

import arrow

from canvas_workflow_kit import events
from canvas_workflow_kit.protocol import (
    STATUS_DUE,
    STATUS_SATISFIED,
    ClinicalQualityMeasure,
    ExternallyAwareClinicalQualityMeasure,
    ProtocolResult
)
from canvas_workflow_kit.recommendation import (
    PerformRecommendation,
    ReferRecommendation,
    TaskRecommendation
)
from canvas_workflow_kit.timeframe import Timeframe
# flake8: noqa
from canvas_workflow_kit.value_set.v2018 import (
    Diabetes,
    OphthalmologicalServices,
    RetinalOrDilatedEyeExam
)
from canvas_workflow_kit.value_set.value_set import ValueSet
from .diabetes_quality_measure import DiabetesQualityMeasure


class FundusPhotography(ValueSet):

    CPT = {'92250'}


class ClinicalQualityMeasure131v6(ExternallyAwareClinicalQualityMeasure, DiabetesQualityMeasure):
    """
    Diabetes: Eye Exam

    Description: Percentage of patients 18-75 years of age with diabetes who had a retinal or
    dilated eye exam by an eye care professional during the measurement period or a negative
    retinal exam (no evidence of retinopathy) in the 12 months prior to the measurement period

    Definition: None

    Rationale: As the seventh leading cause of death in the U.S., diabetes kills approximately
    75,000 people a year (CDC FastStats 2015). Diabetes is a group of diseases marked by high blood
    glucose levels, resulting from the body's inability to produce or use insulin (CDC Statistics
    2014, ADA Basics 2013). People with diabetes are at increased risk of serious health
    complications including vision loss, heart disease, stroke, kidney failure, amputation of toes,
    feet or legs, and premature death. (CDC Fact Sheet 2014).

    In 2012, diabetes cost the U.S. an estimated $245 billion: $176 billion in direct medical costs
    and $69 billion in reduced productivity. This is a 41 percent increase from the estimated $174
    billion spent on diabetes in 2007 (ADA Economic 2013).

    In 2005-2008, of adults with diabetes aged 40 years or older, 4.2 million (28.5%) people had
    diabetic retinopathy, damage to the small blood vessels in the retina that may result in loss
    of vision. (CDC Statistics, 2014).

    Guidance: Only patients with a diagnosis of Type 1 or Type 2 diabetes should be included in the
    denominator of this measure; patients with a diagnosis of secondary diabetes due to another
    condition should not be included.

    The eye exam must be performed by an ophthalmologist or optometrist.

    More information: https://ecqi.healthit.gov/sites/default/files/ecqm/measures/CMS131v6.html
    """

    class Meta:

        title = 'Diabetes: Eye Exam'

        version = '2019-08-02v1'

        description = ('Patients 18-75 years of age with diabetes who have not had a retinal or '
                       'dilated eye exam by an eye care professional.')
        information = 'https://ecqi.healthit.gov/sites/default/files/ecqm/measures/CMS131v6.html'

        identifiers = ['CMS131v6']

        types = ['CQM']

        responds_to_event_types = [
            events.HEALTH_MAINTENANCE,
        ]

        authors = [
            'National Committee for Quality Assurance',
        ]

        references = [
            'American Diabetes Association. Microvascular complications and foot care. Sec. 10. In Standards of Medical Care in Diabetes 2017. Diabetes Care 2017;40(Suppl. 1):S88-S98',
            'American Diabetes Association. 2009. "Standards of Medical Care in Diabetes-2009." Diabetes Care 2009 32 (Suppl 1):S6-S12. doi:10.2337/dc09-S006.',
            'American Diabetes Association. 2013. Diabetes Basics. www.diabetes.org/diabetes-basics/?loc=GlobalNavDB',
            'American Diabetes Association (ADA). April 2013. Economic Costs of Diabetes in the U.S. in 2012. Diabetes Care. Vol. 36 no. 4 1033-46. http://care.diabetesjournals.org/content/36/4/1033.full',
            'Centers for Disease Control and Prevention (CDC). 2014. National Diabetes Statistics Report. http://www.cdc.gov/diabetes/pdfs/data/2014-report-estimates-of-diabetes-and-its-burden-in-the-united-states.pdf',
            'Centers for Disease Control and Prevention (CDC). 2015. FastStats: Deaths and Mortality. www.cdc.gov/nchs/fastats/deaths.htm.',
            'Centers for Disease Control and Prevention. 2014. CDC Features. Diabetes Latest. www.cdc.gov/features/diabetesfactsheet/.', ]

        compute_on_change_types = [
            ClinicalQualityMeasure.CHANGE_PROTOCOL_OVERRIDE,
            ClinicalQualityMeasure.CHANGE_CONDITION,
            ClinicalQualityMeasure.CHANGE_PATIENT,
            ClinicalQualityMeasure.CHANGE_REFERRAL_REPORT,
        ]

    @classmethod
    def enabled(cls) -> bool:
        return True

    _on_date = None
    _in_period = None
    _finding = None

    @property
    def specific_visits(self) -> Type[ValueSet]:
        return super().specific_visits | OphthalmologicalServices

    def in_denominator(self) -> bool:
        """
        Denominator: Equals Initial Population

        Exclusions: Exclude patients who were in hospice care during the measurement year

        Exceptions: None
        """
        if not self.in_initial_population():
            return False

        if self.patient.hospice_within(self.timeframe):
            return False

        return True

    def in_numerator(self) -> bool:
        """
        Numerator: Patients with an eye screening for diabetic retinal disease. This includes
        diabetics who had one of the following:
        A retinal or dilated eye exam by an eye care professional in the measurement period or a
        negative retinal exam (no evidence of retinopathy) by an eye care professional in the year
        prior to the measurement period

        Exclusions: Not Applicable
        """
        records = (self.patient.referral_reports
                   .find(RetinalOrDilatedEyeExam)
                   .within(self.timeframe))  # yapf: disable

        if records:
            self._in_period = True
        else:
            self._in_period = False

            prior_period = Timeframe(
                start=self.timeframe.start.shift(days=-1 * self.timeframe.duration),
                end=self.timeframe.start)

            records = (self.patient.referral_reports
                       .find(RetinalOrDilatedEyeExam)
                       .within(prior_period))  # yapf: disable

        if records:
            last = records[-1]
            negative_finding = False
            self._finding = ''

            for coding in last['codings']:
                if coding['display'] == 'Findings':
                    self._finding = coding['value']

                if coding['code'] == '721103006':
                    negative_finding = True

            self._on_date = arrow.get(last['originalDate'])

            if self._in_period or negative_finding:
                return True

        return False

    # protocol satisfied
    def craft_satisfied_result(self):
        result = ProtocolResult()
        result.status = STATUS_SATISFIED
        exam_date = self.display_date(self._on_date)

        if self._in_period:
            result.due_in = (self._on_date.shift(days=self.timeframe.duration) - self.now).days
            result.add_narrative(
                f'{self.patient.first_name} has diabetes and a retinal examination was done '
                f'{exam_date}, demonstrating {self._finding}.')
        else:
            result.due_in = (self._on_date.shift(days=self.timeframe.duration) - self.now).days
            next_date = self._on_date.shift(days=self.timeframe.duration).format('M/D/YY')
            result.add_narrative(
                f'{self.patient.first_name} has diabetes and a retinal examination was done '
                f'{exam_date} demonstrating no diabetic eye disease. '
                f'Next examination is due {next_date}.')

        return result

    def craft_unsatisfied_result(self):
        result = ProtocolResult()
        result.due_in = -1
        result.status = STATUS_DUE
        if self._on_date:
            exam_date = self.display_date(self._on_date)
            result.add_narrative(
                f'{self.patient.first_name} has diabetes and a prior abnormal retinal examination '
                f'{exam_date} showing {self._finding}. {self.patient.first_name} is due for '
                'retinal examination.')
        else:
            humanized_timeframe = ((self.now
                                    .shift(months=-1,
                                           days=-self.timeframe.duration)
                                    .humanize(other=self.now,
                                              granularity='month',
                                              only_distance=True))
                                   .replace(' ago', ''))  # yapf: disable
            result.add_narrative(
                f'{self.patient.first_name} has diabetes and no documentation of retinal '
                f'examination in the past {humanized_timeframe}.')

        result.add_recommendation(
            PerformRecommendation(
                key='CMS131v6_RECOMMEND_PERFORM_EYE_EXAM',
                rank=1,
                button='Perform',
                patient=self.patient,
                procedure=FundusPhotography,
                condition=Diabetes,
                title='Perform retinal examination'))

        context_conditions = (self.patient.conditions.find(Diabetes).intersects(
            self.timeframe, still_active=self.patient.active_only))

        codings = [condition['coding'] for condition in context_conditions.records]

        result.add_recommendation(
            ReferRecommendation(
                key='CMS131v6_RECOMMEND_REFER_EYE_EXAM',
                rank=2,
                button='Refer',
                patient=self.patient,
                referral=RetinalOrDilatedEyeExam,
                condition=Diabetes,
                context={
                    'conditions': codings,
                    'specialties': [
                        'Ophthalmology',
                        'Ophthalmology ',  # TODO/BUG: science contacts contains an entry like this
                        'Optometrist',
                        'Optometry',
                    ],
                },
                title='Refer for retinal examination'))

        return result



from protocols import events
from protocols.protocol import ClinicalQualityMeasure
from protocols.value_set.v2021 import (
    Cancer, Ethnicity, OfficeVisit, OncAdministrativeSex, Payer, Race,
    StandardizedPainAssessmentTool)


class ClinicalQualityMeasure157v9(ClinicalQualityMeasure):
    """
    Oncology: Medical and Radiation - Pain Intensity Quantified

    Description: Percentage of patient visits, regardless of patient age, with a diagnosis of
    cancer currently receiving chemotherapy or radiation therapy in which pain intensity is
    quantified

    Definition: None

    Rationale: An estimated 1.7 million new cases of cancer are diagnosed in the US each year (NIH,
    2017). Pain is a commonly occurring symptom for cancer patients as 30% to 50% (510,000 to
    850,000 each year based on current statistics) will experience moderate to severe pain (Wiffen,
    Wee, Derry, Bell, & Moore, 2017). Initial and ongoing pain assessments are essential to
    determine the pathophysiology of pain and ensure proper pain management. According to the
    National Comprehensive Cancer Network (NCCN), there is increasing evidence in oncology that
    survival is linked to symptom reporting and control and that pain management contributes to
    broad quality-of-life improvement (2018). Cancer patients have reported that pain interferes
    with their mood, work, relationships with other people, sleep and overall enjoyment of life
    (Moryl et al., 2018). To maximize patient outcomes, pain management is an essential part of
    oncologic management (NCCN, 2018).

    A recent analysis of registry data for chronic pain cancer patients found average pain
    intensity reported as mild (24.6% of patients), moderate (41.5%), and severe (33.9%). The study
    also indicated that patient report of pain relief is inversely related to the average pain
    intensity reported (Moryl et al., 2018). These data suggest that assessing and managing a
    cancer patient’s pain is critical and there remains significant room for improvement in
    assessing and mitigating cancer-related pain. A prospective study of changes in pain severity
    of cancer patients found that, at initial assessment, 47% of patients reported pain. At follow-
    up, the patients with pain at initial assessment reported reduced pain (32.2%), stable pain
    (48.2%) and worse pain (19.6%). Of the 53% of patients reporting no pain at initial assessment,
    82.6% reported stable pain and 17.4% reported worse pain at follow-up assessment (Zhao et al.,
    2014). This study highlights the importance of initial and ongoing assessments of pain to
    identify gaps and ensure proper pain management.

    Guidance: This eCQM is an episode-of-care measure; the level of analysis for this measure is
    every visit for patients with a diagnosis of cancer who are also currently receiving
    chemotherapy or radiation therapy during the measurement period.

    For patients receiving radiation therapy, pain intensity should be quantified at each radiation
    treatment management encounter where the patient and physician have a face-to-face interaction.
    Due to the nature of some applicable coding related to radiation therapy (e.g., delivered in
    multiple fractions), the billing date for certain codes may or may not be the same as the face-
    to-face encounter date. In this instance, for the reporting purposes of this measure, the
    billing date should be used to pull the appropriate patients into the initial population. It is
    expected, though, that the numerator criteria would be performed at the time of the actual
    face-to-face encounter during the series of treatments. A lookback (retrospective) period of 7
    days, including the billing date, may be used to identify the actual face-to-face encounter,
    which is required to assess the numerator. Therefore, pain intensity should be quantified
    during the face-to-face encounter occurring on the actual billing date or within the 6 days
    prior to the billing date.

    For patients receiving chemotherapy, pain intensity should be quantified at each face-to-face
    encounter with the physician while the patient is currently receiving chemotherapy. For
    purposes of identifying eligible encounters, patients "currently receiving chemotherapy" refers
    to patients administered chemotherapy within 30 days prior to the encounter AND administered
    chemotherapy within 30 days after the date of the encounter.

    Pain intensity should be quantified using a standard instrument, such as a 0-10 numerical
    rating scale, visual analog scale, a categorical scale, or pictorial scale. Examples include
    the Faces Pain Rating Scale and the Brief Pain Inventory (BPI).

    This version of the eCQM uses QDM version 5.5. Please refer to the eCQI resource center
    (https://ecqi.healthit.gov/qdm) for more information on the QDM.

    More information: https://ecqi.healthit.gov/ecqm/measures/cms157v9
    """

    title = 'Oncology: Medical and Radiation - Pain Intensity Quantified'

    identifiers = ['CMS157v9']

    types = ['CQM']

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'American Medical Association (AMA)',
        'PCPI(R) Foundation (PCPI[R])',
    ]

    references = [
        'Moryl, N., Dave, V., Glare, P., Bokhari, A., Malhotra, V. T., Gulati, A., … Inturrisi, C. E. (2018). Patient-Reported Outcomes and Opioid Use by Outpatient Cancer Patients. The Journal of Pain: official journal of the American Pain Society, 19(3), 278–290. doi:10.1016/j.jpain.2017.11.001',
        'National Comprehensive Cancer Network (NCCN). (2018). NCCN Clinical Practice Guidelines in Oncology. Adult Cancer Pain Version I.2018. Retrieved from http://www.nccn.org',
        'National Comprehensive Cancer Network (NCCN). (2019). NCCN Clinical Practice Guidelines in Oncology. Adult Cancer Pain Version 3.2019. Retrieved from http://www.nccn.org',
        'National Institutes of Health - National Cancer Institute. (2017). Cancer Statistics. Retrieved from https://www.cancer.gov/about-cancer/understanding/statistics',
        'Wiffen, P. J., Wee, B., Derry, S., Bell, R. F., & Moore, R. A. (2017). Opioids for cancer pain - an overview of Cochrane reviews. The Cochrane database of systematic reviews, 7(7), CD012592. doi:10.1002/14651858.CD012592.pub2',
        'Zhao, F., Chang, V. T., Cleeland, C., Cleary, J. F., Mitchell, E. P., Wagner, L. I., & Fisch, M. J. (2014). Determinants of pain severity changes in ambulatory patients with cancer: an analysis from Eastern Cooperative Oncology Group trial E2Z02. Journal of clinical oncology : official journal of the American Society of Clinical Oncology, 32(4), 312–319. doi:10.1200/JCO.2013.50.6071',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: All patient visits, regardless of patient age, with a diagnosis of
        cancer currently receiving chemotherapy or radiation therapy
        """
        pass

    def in_denominator(self):
        """
        Denominator: Equals Initial Population

        Exclusions: None

        Exceptions: None
        """
        return self.in_initial_population()

    def in_numerator(self):
        """
        Numerator: Patient visits in which pain intensity is quantified

        Exclusions: Not Applicable
        """
        pass


    def compute_results(self):
        """
        Clinical recommendation: -Screen all patients for pain at each contact.

        -Routinely quantify and document pain intensity and quality as characterized by the patient
        (whenever possible). Include patient reporting of breakthrough pain, treatments used and
        their impact on pain, adequacy of comfort, satisfaction with pain relief, provider
        assessment of impact on function, and any special issues for the patient relevant to pain
        treatment. If necessary, get additional information from caregiver regarding pain and
        impact on function.

        -Perform comprehensive pain assessment if new or worsening pain is present, and regularly
        for persisting pain.

        Various methods and tools exist to assess pain severity. Intensity of pain should be
        quantified using a numerical rating scale (i.e., 0-10), visual analog scale, categorical
        scale, or pictorial scale (e.g., The Faces Pain Rating Scale) (Category 2A) (National
        Comprehensive Cancer Network, 2019).
        """
        pass



from ..value_set import ValueSet


class NegativeDepressionScreening(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent situations in which a depression screen was performed by the health care professional and the finding was negative for depression.

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) category related to Assessment.

    **Inclusion Criteria:** Includes only relevant concepts associated with screening for depression whereby the outcome is negative.

    **Exclusion Criteria:** No exclusions.
    """

    OID = '2.16.840.1.113883.3.526.3.1564'
    VALUE_SET_NAME = 'Negative Depression Screening'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    SNOMEDCT = {
        '428171000124102'
    }

class PositiveDepressionScreening(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent situations in which a depression screen was performed by the health care professional and the finding or result was positive for depression.

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) category related to Assessment.

    **Inclusion Criteria:** Includes only relevant concepts associated with screening for depression whereby the outcome is positive.

    **Exclusion Criteria:** No exclusions.
    """

    OID = '2.16.840.1.113883.3.526.3.1565'
    VALUE_SET_NAME = 'Positive Depression Screening'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    SNOMEDCT = {
        '428181000124104'
    }

class StandardizedPainAssessmentTool(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent pain assessment tools or instruments used to quantify pain intensity.

    **Data Element Scope:** This value set may use Quality Data Model (QDM) category related to Assessment.

    **Inclusion Criteria:** Includes only relevant concepts associated with pain assessment tools or instruments used to quantify pain intensity such as 0-10 numerical rating scale, visual analog scale, a categorical scale, pictorial scale, faces pain rating scale, or the Brief Pain Inventory (BPI).

    **Exclusion Criteria:** No exclusions.
    """

    OID = '2.16.840.1.113883.3.526.3.1028'
    VALUE_SET_NAME = 'Standardized Pain Assessment Tool'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    LOINC = {
        '38208-5',
        '38214-3',
        '38221-8',
        '72514-3',
        '77565-0'
    }

class StandardizedToolsForAssessmentOfCognition(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent the total score results for the standardized tools used for the assessment of cognition.

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) category related to Assessment.

    **Inclusion Criteria:** Includes only relevant concepts associated with total score results for the following standardized tools: -Blessed Orientation-Memory-Concentration Test (BOMC) -Montreal Cognitive Assessment (MoCA) -St. Louis University Mental Status Examination (SLUMS) -Mini-Mental State Examination (MMSE) [Note: The MMSE has not been well validated for non-Alzheimer's dementias] -Short Informant Questionnaire on Cognitive Decline in the Elderly (IQCODE) -Ascertain Dementia 8 (AD8) Questionnaire -Minimum Data Set (MDS) Brief Interview of Mental Status (BIMS) [Note: Validated for use with nursing home patients only] -Formal neuropsychological evaluation -Mini-Cog.

    **Exclusion Criteria:** No exclusions.
    """

    OID = '2.16.840.1.113883.3.526.3.1006'
    VALUE_SET_NAME = 'Standardized Tools for Assessment of Cognition'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    LOINC = {
        '58151-2',
        '71492-3',
        '71493-1',
        '71722-3',
        '72106-8',
        '72172-0',
        '72173-8',
        '72233-0'
    }

class TobaccoUseScreening(ValueSet):
    """
    **Clinical Focus:** This value set contains concepts that represent screening for tobacco use.

    **Data Element Scope:** This value set may use the Quality Data Model (QDM) category related to Assessment.

    **Inclusion Criteria:** Includes only relevant concepts associated with screening tools or questions used to obtain a patient's tobacco use status.

    **Exclusion Criteria:** No exclusions.
    """

    OID = '2.16.840.1.113883.3.526.3.1278'
    VALUE_SET_NAME = 'Tobacco Use Screening'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    LOINC = {
        '39240-7',
        '68535-4',
        '68536-2',
        '72166-2'
    }

class UrinaryRetention(ValueSet):
    """
    **Clinical Focus:** This set of values focuses on urinary retention related to drugs or other causes.

    **Data Element Scope:** The intent of this data element is to define urinary retention.

    **Inclusion Criteria:** Urinary retention related to drugs or other causes

    **Exclusion Criteria:** None
    """

    OID = '2.16.840.1.113762.1.4.1164.52'
    VALUE_SET_NAME = 'Urinary retention'
    EXPANSION_VERSION = 'eCQM Update 2020-05-07'

    ICD10CM = {
        'R330',
        'R338',
        'R339'
    }

    SNOMEDCT = {
        '12245681000119103',
        '236648008',
        '267064002'
    }


"""

Events are specified to allow the system to know when to compute the protocol.

They are declared in the `responds_to_event_types` class attribute in `ClinicalQualityMeasure`
classes.


```python
from canvas_workflow_kit import events

class Hcc002v2(ClinicalQualityMeasure):
    title = 'CKD suspect'

    ...

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    ...
```


## Event triggers

### HEALTH_MAINTENANCE
**canvas\_workflow\_kit.events.HEALTH\_MAINTENANCE**

Run protocol when patient data has changed.

### CHART_OPEN
**canvas\_workflow\_kit.events.CHART\_OPEN**

Run protocol when patient the patient chart has been opened.

"""

# Run protocol when patient data has changed.
HEALTH_MAINTENANCE = 'HEALTH_MAINTENANCE'

# Run protocol when patient the patient chart has been opened.
CHART_OPEN = 'CHART_OPEN'



from ..value_set import ValueSet


class RetinalOrDilatedEyeExam(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for a physical exam where a retinal or dilated eye exam was performed.

    **Data Element Scope:** This value set may use a model element related to Physical Exam.

    **Inclusion Criteria:** Includes concepts that represent a physical exam during which a retinal or dilated eye exam occurred.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS131v10
    """

    VALUE_SET_NAME = 'Retinal or Dilated Eye Exam'
    OID = '2.16.840.1.113883.3.464.1003.115.12.1088'
    DEFINITION_VERSION = '20170504'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    SNOMEDCT = {
        '6615001',  # Electroretinography (procedure)
        '252779009',  # Single bright white flash electroretinography (procedure)
        '252780007',  # Dark adapted single bright flash electroretinography (procedure)
        '252781006',  # Pre-dark-adapted single bright flash electroretinography (procedure)
        '252782004',  # Photopic electroretinography (procedure)
        '252783009',  # Scotopic rod electroretinography (procedure)
        '252784003',  # Flicker electroretinography (procedure)
        '252788000',  # Chromatic electroretinography (procedure)
        '252789008',  # Early receptor potential electroretinography (procedure)
        '252790004',  # Focal electroretinography (procedure)
        '274795007',  # Examination of optic disc (procedure)
        '274798009',  # Examination of retina (procedure)
        '308110009',  # Direct fundoscopy following mydriatic (procedure)
        '314971001',  # Camera fundoscopy (procedure)
        '314972008',  # Indirect fundoscopy following mydriatic (procedure)
        '410451008',  # Indirect ophthalmoscopy (procedure)
        '410452001',  # Monocular indirect ophthalmoscopy (procedure)
        '410453006',  # Binocular indirect ophthalmoscopy (procedure)
        '410455004',  # Slit lamp fundus examination (procedure)
        '420213007',  # Multifocal electroretinography (procedure)
        '425816006',  # Ultrasonic evaluation of retina (procedure)
        '427478009',  # Evaluation of retina (procedure)
        '722161008',  # Diabetic retinal eye exam (procedure)
    }


class BestCorrectedVisualAcuityExamUsingSnellenChart(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts of a physical exam for visual acuity exams using a Snellen chart.

    **Data Element Scope:** This value set may use a model element related to Physical Exam.

    **Inclusion Criteria:** Includes concepts that represent a physical exam for visual acuity using a Snellen chart.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS133v10
    """

    VALUE_SET_NAME = 'Best Corrected Visual Acuity Exam Using Snellen Chart'
    OID = '2.16.840.1.113883.3.526.3.1560'
    DEFINITION_VERSION = '20210210'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    LOINC = {
        '79880-1',  # Visual acuity best corrected Right eye by Snellen eye chart
        '79881-9',  # Visual acuity best corrected Left eye by Snellen eye chart
    }


class HeartRate(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts of a physical exam measuring heart rate.

    **Data Element Scope:** This value set may use a model element related to Physical Exam.

    **Inclusion Criteria:** Includes concepts that represent a physical exam for obtaining heart rate.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS144v10, CMS145v10
    """

    VALUE_SET_NAME = 'Heart Rate'
    OID = '2.16.840.1.113883.3.526.3.1176'
    DEFINITION_VERSION = '20170504'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    LOINC = {
        '68999-2',  # Heart rate --supine
        '69000-8',  # Heart rate --sitting
        '69001-6',  # Heart rate --standing
        '8867-4',  # Heart rate
    }


class BmiPercentile(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for a physical exam where a body mass index (BMI) percentile is calculated.

    **Data Element Scope:** This value set may use a model element related to Physical Exam.

    **Inclusion Criteria:** Includes concepts that represent a physical exam with a BMI percentile measurement.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS155v10
    """

    VALUE_SET_NAME = 'BMI percentile'
    OID = '2.16.840.1.113883.3.464.1003.121.12.1012'
    DEFINITION_VERSION = '20170504'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    LOINC = {
        '59574-4',  # Body mass index (BMI) [Percentile]
        '59575-1',  # Body mass index (BMI) [Percentile] Per age
        '59576-9',  # Body mass index (BMI) [Percentile] Per age and sex
    }


class Height(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for a physical exam with measured height.

    **Data Element Scope:** This value set may use a model element related to Physical Exam.

    **Inclusion Criteria:** Includes concepts that represent a physical exam with a patient height measurement.

    **Exclusion Criteria:** Excludes concepts that represent patient reported height, a percentile for height or an estimated height.

    ** Used in:** CMS155v10
    """

    VALUE_SET_NAME = 'Height'
    OID = '2.16.840.1.113883.3.464.1003.121.12.1014'
    DEFINITION_VERSION = '20170504'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    LOINC = {
        '3137-7',  # Body height Measured
        '8302-2',  # Body height
        '8306-3',  # Body height --lying
        '8307-1',  # Body height --preoperative
        '8308-9',  # Body height --standing
    }


class Weight(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for a physical exam with measured weight.

    **Data Element Scope:** This value set may use a model element related to Physical Exam.

    **Inclusion Criteria:** Includes concepts that represent a physical exam with a patient weight measurement.

    **Exclusion Criteria:** Excludes concepts that represent an ideal body weight, estimated body weight or a body fat measurement.

    ** Used in:** CMS155v10
    """

    VALUE_SET_NAME = 'Weight'
    OID = '2.16.840.1.113883.3.464.1003.121.12.1015'
    DEFINITION_VERSION = '20170504'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    LOINC = {
        '18833-4',  # First Body weight
        '29463-7',  # Body weight
        '3141-9',  # Body weight Measured
        '3142-7',  # Body weight Stated
        '8341-0',  # Dry body weight Measured
        '8349-3',  # Body weight Measured --preoperative
        '8350-1',  # Body weight Measured --with clothes
        '8351-9',  # Body weight Measured --without clothes
    }


class BmiRatio(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for a physical exam where a body mass index (BMI) ratio is calculated.

    **Data Element Scope:** This value set may use a model element related to Physical Exam.

    **Inclusion Criteria:** Includes concepts that represent a physical exam with a body mass index (BMI) ratio measurement.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS249v4
    """

    VALUE_SET_NAME = 'BMI Ratio'
    OID = '2.16.840.1.113883.3.600.1.1490'
    DEFINITION_VERSION = '20140502'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    LOINC = {
        '39156-5',  # Body mass index (BMI) [Ratio]
    }




from protocols import events
from protocols.protocol import ClinicalQualityMeasure
from protocols.value_set.v2021 import (
    Ethnicity, GroupPsychotherapy, OfficeVisit, OncAdministrativeSex, OutpatientConsultation,
    Payer, PsychVisitDiagnosticEvaluation, PsychVisitFamilyPsychotherapy, PsychVisitPsychotherapy,
    Psychoanalysis, Race, TelehealthServices)


class ClinicalQualityMeasure177v9(ClinicalQualityMeasure):
    """
    Child and Adolescent Major Depressive Disorder (MDD): Suicide Risk Assessment

    Description: Percentage of patient visits for those patients aged 6 through 17 years with a
    diagnosis of major depressive disorder with an assessment for suicide risk

    Definition: Numerator Definition: The specific type and magnitude of the suicide risk
    assessment is intended to be at the discretion of the individual clinician and should be
    specific to the needs of the patient. At a minimum, suicide risk assessment should evaluate:
    1. Risk (e.g., age, sex, stressors, comorbid conditions, hopelessness, impulsivity) and
    protective factors (e.g., religious belief, concern not to hurt family) that may influence the
    desire to attempt suicide.
    2. Current severity of suicidality.
    3. Most severe point of suicidality in episode and lifetime.

    Low burden tools to track suicidal ideation and behavior such as the Columbia-Suicidal Severity
    Rating Scale can also be used. Because no validated assessment tool or instrument fully meets
    the aforementioned requirements for the suicide risk assessment, individual tools or
    instruments have not been explicitly included in coding.

    Rationale: Research has shown that patients with major depressive disorder are at a high risk
    for suicide attempts and completion - among the most significant and devastating sequelae of
    the disease. Suicide risk is a critical consideration in children and adolescents with MDD and
    an important aspect of care that should be assessed at each visit and subsequently managed to
    minimize that risk. Additionally, the importance of the assessments is underscored by research
    (Louma, Martin, & Pearson, 2002) that indicates that many individuals who die by suicide do
    make contact with primary care providers and mental health services beforehand. More
    specifically, approximately 15% of suicide victims aged 35 years or younger had seen a mental
    health professional within 1 month of suicide while approximately 23% had seen a primary care
    provider within 1 month of suicide.

    Guidance: This eCQM is an episode-based measure. A suicide risk assessment should be performed
    at every visit for major depressive disorder during the measurement period.

    In recognition of the growing use of integrated and team-based care, the diagnosis of
    depression and the assessment for suicide risk need not be performed by the same provider or
    clinician.

    Suicide risk assessments completed via telehealth services can also meet numerator performance.

    This measure is an episode-of-care measure; the level of analysis for this measure is every
    visit for major depressive disorder during the measurement period. For example, at every visit
    for MDD, the patient should have a suicide risk assessment.

    Use of a standardized tool(s) or instrument(s) to assess suicide risk will meet numerator
    performance, so long as the minimum criteria noted above is evaluated. Standardized tools can
    be mapped to the concept "Intervention, Performed": "Suicide risk assessment (procedure)"
    included in the numerator logic below, as no individual suicide risk assessment tool or
    instrument would satisfy the requirements alone.

    This version of the eCQM uses QDM version 5.5.  Please refer to the eCQI resource center
    (https://ecqi.healthit.gov/qdm) for more information on the QDM.

    More information: https://ecqi.healthit.gov/ecqm/measures/cms177v9
    """

    title = 'Child and Adolescent Major Depressive Disorder (MDD): Suicide Risk Assessment'

    identifiers = ['CMS177v9']

    types = ['CQM']

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'American Medical Association (AMA)',
        'PCPI(R) Foundation (PCPI[R])',
    ]

    references = [
        'American Academy of Child and Adolescent Psychiatry. (2007). Practice parameter for the assessment and treatment of children and adolescents with depressive disorders. Journal of the American Academy of Child and Adolescent Psychiatry, 46(11), 1503-1526. doi:10.1097/chi.0b013e318145ae1c',
        'American Psychiatric Association. (2010). Practice guideline for the treatment of patients with major depressive disorder. 3rd edition. Retrieved from http://psychiatryonline.org/pb/assets/raw/sitewide/practice_guidelines/guidelines/mdd.pdf (This guideline was reaffirmed in 2015.)',
        'Luoma, J. B., Martin, C. E., & Pearson, J. L. (2002). Contact with mental health and primary care providers before suicide: A review of the evidence. American Journal of Psychiatry, 159(6), 909-916. doi:10.1176/appi.ajp.159.6.909',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: All patient visits for those patients aged 6 through 17 years with a
        diagnosis of major depressive disorder
        """
        pass

    def in_denominator(self):
        """
        Denominator: Equals Initial Population

        Exclusions: None

        Exceptions: None
        """
        return self.in_initial_population()

    def in_numerator(self):
        """
        Numerator: Patient visits with an assessment for suicide risk

        Exclusions: Not Applicable
        """
        pass


    def compute_results(self):
        """
        Clinical recommendation: The evaluation must include assessment for the presence of harm to
        self or others (MS) (American Academy of Child and Adolescent Psychiatry, 2007).

        Suicidal behavior exists along a continuum from passive thoughts of death to a clearly
        developed plan and intent to carry out that plan. Because depression is closely associated
        with suicidal thoughts and behavior, it is imperative to evaluate these symptoms at the
        initial and subsequent assessments. For this purpose, low burden tools to track suicidal
        ideation and behavior such as the Columbia-Suicidal Severity Rating Scale can be used.
        Also, it is crucial to evaluate the risk (e.g., age, sex, stressors, comorbid conditions,
        hopelessness, impulsivity) and protective factors (e.g., religious belief, concern not to
        hurt family) that might influence the desire to attempt suicide. The risk for suicidal
        behavior increases if there is a history of suicide attempts, comorbid psychiatric
        disorders (e.g., disruptive disorders, substance abuse), impulsivity and aggression,
        availability of lethal agents (e.g., firearms), exposure to negative events (e.g., physical
        or sexual abuse, violence), and a family history of suicidal behavior (American Academy of
        Child and Adolescent Psychiatry, 2007).

        A careful and ongoing evaluation of suicide risk is necessary for all patients with major
        depressive disorder (Category I). Such an assessment includes specific inquiry about
        suicidal thoughts, intent, plans, means, and behaviors; identification of specific
        psychiatric symptoms (e.g., psychosis, severe anxiety, substance use) or general medical
        conditions that may increase the likelihood of acting on suicidal ideas; assessment of past
        and, particularly, recent suicidal behavior; delineation of current stressors and potential
        protective factors (e.g., positive reasons for living, strong social support); and
        identification of any family history of suicide or mental illness (Category I) (American
        Psychiatric Association, 2010, reaffirmed 2015).
        """
        pass



from ..value_set import ValueSet


class LevelOfSeverityOfRetinopathyFindings(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for diagnoses of the level of severity of retinopathy.

    **Data Element Scope:** This value set may use a model element related to Diagnosis or Communication.

    **Inclusion Criteria:** Includes concepts that represent a diagnosis of mild, moderate, severe, or very severe non-proliferative and proliferative diabetic retinopathy.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS142v10
    """

    VALUE_SET_NAME = 'Level of Severity of Retinopathy Findings'
    OID = '2.16.840.1.113883.3.526.3.1283'
    DEFINITION_VERSION = '20210209'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    SNOMEDCT = {
        '59276001',  # Proliferative retinopathy due to diabetes mellitus (disorder)
        '312903003',  # Mild nonproliferative retinopathy due to diabetes mellitus (disorder)
        '312904009',  # Moderate nonproliferative retinopathy due to diabetes mellitus (disorder)
        '312905005',  # Severe nonproliferative retinopathy due to diabetes mellitus (disorder)
        '399876000',  # Very severe nonproliferative retinopathy due to diabetes mellitus (disorder)
    }


class MacularEdemaFindingsPresent(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for diagnoses to identify the presence of macular edema.

    **Data Element Scope:** This value set may use a model element related to Diagnosis.

    **Inclusion Criteria:** Includes concepts that represent a diagnosis of macular edema.

    **Exclusion Criteria:** Excludes concepts that represent a diagnosis of non-macular edema.

    ** Used in:** CMS142v10
    """

    VALUE_SET_NAME = 'Macular Edema Findings Present'
    OID = '2.16.840.1.113883.3.526.3.1320'
    DEFINITION_VERSION = '20210209'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    SNOMEDCT = {
        '37231002',  # Macular retinal edema (disorder)
        '193350004',  # Advanced maculopathy due to diabetes mellitus (disorder)
        '193387007',  # Cystoid macular edema (disorder)
        '232020009',  # Disorder of macula due to diabetes mellitus (disorder)
        '312911008',  # Clinically significant macular edema (disorder)
        '312912001',  # Macular edema due to diabetes mellitus (disorder)
        '312920004',  # Postoperative cystoid macular edema (disorder)
        '312921000',  # Autosomal dominant cystoid macular edema (disorder)
        '312922007',  # Uveitis related cystoid macular edema (disorder)
        '314010006',  # Diffuse exudative maculopathy due to diabetes mellitus (disorder)
        '314011005',  # Focal exudative maculopathy due to diabetes mellitus (disorder)
        '314014002',  # Ischemic maculopathy due to diabetes mellitus (disorder)
        '314015001',  # Mixed maculopathy due to diabetes mellitus (disorder)
        '399864000',  # Macular edema not clinically significant due to diabetes mellitus (disorder)
        '399874002',  # High risk proliferative retinopathy with clinically significant macula edema due to diabetes mellitus (disorder)
        '399875001',  # Non-high-risk proliferative retinopathy with clinically significant macular edema due to diabetes mellitus (disorder)
        '420486006',  # Exudative maculopathy due to type 1 diabetes mellitus (disorder)
        '421779007',  # Exudative maculopathy due to type 2 diabetes mellitus (disorder)
        '432789001',  # Noncystoid edema of macula of retina (disorder)
        '769217008',  # Macular edema of right eye due to diabetes mellitus (disorder)
        '769218003',  # Macular edema of left eye due to diabetes mellitus (disorder)
        '769219006',  # Macular edema due to type 1 diabetes mellitus (disorder)
        '769220000',  # Macular edema due to type 2 diabetes mellitus (disorder)
        '770097006',  # Clinically significant macular edema due to diabetes mellitus (disorder)
        '870529009',  # Persistent macular edema due to diabetes mellitus (disorder)
        '871778008',  # Centrally involved macular edema due to diabetes mellitus (disorder)
        '871781003',  # Non centrally involved macular edema due to diabetes mellitus (disorder)
    }


class ConsultantReport(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts for consultant reports.

    **Data Element Scope:** This value set may use a model element related to Assessment.

    **Inclusion Criteria:** Includes concepts that represent various consultant reports and notes.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS50v10
    """

    VALUE_SET_NAME = 'Consultant Report'
    OID = '2.16.840.1.113883.3.464.1003.121.12.1006'
    DEFINITION_VERSION = '20190315'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    LOINC = {
        '11488-4',  # Consult note
        '34099-2',  # Cardiology Consult note
        '34100-8',  # Intensive care unit Consult note
        '34101-6',  # General medicine Outpatient Consult note
        '34102-4',  # Psychiatry Hospital Consult note
        '34103-2',  # Pulmonary Consult note
        '34104-0',  # Hospital Consult note
        '34749-2',  # Anesthesiology Outpatient Consult note
        '34756-7',  # Dentistry Consult note
        '34758-3',  # Dermatology Consult note
        '34760-9',  # Diabetology Consult note
        '34761-7',  # Gastroenterology Consult note
        '34764-1',  # General medicine Consult note
        '34776-5',  # Geriatric medicine Consult note
        '34777-3',  # Obstetrics and Gynecology Consult note
        '34779-9',  # Hematology+Medical oncology Consult note
        '34781-5',  # Infectious disease Consult note
        '34783-1',  # Kinesiotherapy Consult note
        '34785-6',  # Mental health Consult note
        '34788-0',  # Psychiatry Consult note
        '34791-4',  # Psychology Consult note
        '34795-5',  # Nephrology Consult note
        '34797-1',  # Neurology Consult note
        '34798-9',  # Neurological surgery Consult note
        '34800-3',  # Nutrition and dietetics Consult note
        '34803-7',  # Occupational medicine Consult note
        '34805-2',  # Oncology Consult note
        '34807-8',  # Ophthalmology Consult note
        '34810-2',  # Optometry Consult note
        '34812-8',  # Oral and Maxillofacial Surgery Consult note
        '34814-4',  # Orthopaedic surgery Consult note
        '34816-9',  # Otolaryngology Consult note
        '34820-1',  # Pharmacology Consult note
        '34822-7',  # Physical medicine and rehab Consult note
        '34824-3',  # Physical therapy Consult note
        '34826-8',  # Plastic surgery Consult note
        '34828-4',  # Podiatry Consult note
        '34831-8',  # Radiation oncology Consult note
        '34833-4',  # Recreational therapy Consult note
        '34837-5',  # Respiratory therapy Consult note
        '34839-1',  # Rheumatology Consult note
        '34841-7',  # Social worker Consult note
        '34845-8',  # Speech-language pathology+Audiology Consult note
        '34847-4',  # Surgery Consult note
        '34849-0',  # Cardiothoracic surgery Consult note
        '34851-6',  # Urology Consult note
        '34853-2',  # Vascular surgery Consult note
        '34855-7',  # Occupational therapy Consult note
        '34879-7',  # Endocrinology Consult note
        '51845-6',  # Outpatient Consult note
        '51846-4',  # Emergency department Consult note
        '51854-8',  # Long term care facility Consult note
        '64056-5',  # General medicine Medical student Hospital Consult note
        '64068-0',  # Surgery Medical student Hospital Consult note
        '64072-2',  # Critical care medicine Medical student Hospital Consult note
        '64076-3',  # Cardiothoracic surgery Medical student Hospital Consult note
        '64080-5',  # Pulmonary Medical student Hospital Consult note
        '68469-6',  # Pastoral care Hospital Consult note
        '68486-0',  # Cardiology Medical student Hospital Consult note
        '68551-1',  # Dermatology Hospital Consult note
        '68566-9',  # Obstetrics and Gynecology Hospital Consult note
        '68570-1',  # Occupational therapy Hospital Consult note
        '68575-0',  # Ophthalmology Hospital Consult note
        '68586-7',  # Pharmacology Hospital Consult note
        '68590-9',  # Physical therapy Hospital Consult note
        '68597-4',  # Plastic surgery Hospital Consult note
        '68619-6',  # Adolescent medicine Hospital Consult note
        '68633-7',  # Allergy and Immunology Hospital Consult note
        '68639-4',  # Audiology Hospital Consult note
        '68648-5',  # Child and adolescent psychiatry Hospital Consult note
        '68651-9',  # Clinical biochemical genetics Hospital Consult note
        '68661-8',  # Clinical genetics Hospital Consult note
        '68670-9',  # Developmental-behavioral pediatrics Hospital Consult note
        '68681-6',  # Multi-specialty program Hospital Consult note
        '68685-7',  # Neonatal perinatal medicine Hospital Consult note
        '68694-9',  # Neurological surgery Hospital Consult note
        '68705-3',  # Neurology with special qualifications in child neurology Hospital Consult note
        '68716-0',  # Pain medicine Hospital Consult note
        '68727-7',  # Pediatric cardiology Hospital Consult note
        '68746-7',  # Pediatric gastroenterology Hospital Consult note
        '68757-4',  # Pediatric hematology-oncology Hospital Consult note
        '68765-7',  # Pediatric infectious diseases Hospital Consult note
        '68787-1',  # Pediatric pulmonology Hospital Consult note
        '68802-8',  # Pediatric surgery Hospital Consult note
        '68812-7',  # Pediatric urology Hospital Consult note
        '68821-8',  # Pediatrics Hospital Consult note
        '68837-4',  # Primary care Hospital Consult note
        '68846-5',  # Speech-language pathology Hospital Consult note
        '68852-3',  # Transplant surgery Hospital Consult note
        '68864-8',  # Pediatric transplant hepatology Hospital Consult note
        '68869-7',  # Pediatric nephrology Hospital Consult note
        '68874-7',  # Pediatric otolaryngology Hospital Consult note
        '68879-6',  # Pediatric rheumatology Hospital Consult note
        '68892-9',  # Pediatric dermatology Hospital Consult note
        '68897-8',  # Pediatric endocrinology Hospital Consult note
        '72555-6',  # Interventional radiology Consult note
        '73575-3',  # Radiology Consult note
        '75424-2',  # Audiology Consult note
        '75465-5',  # Orthotics prosthetics Consult note
        '77403-4',  # Anesthesiology Consult note
        '77429-9',  # Allergy and Immunology Consult note
        '78250-8',  # Colon and rectal surgery Consult note
        '78251-6',  # Hematology Consult note
        '78252-4',  # Multi-specialty program Consult note
        '78253-2',  # Family medicine Consult note
        '78254-0',  # Clinical genetics Consult note
        '78405-8',  # Neonatal perinatal medicine Consult note
        '78406-6',  # Nurse practitioner Consult note
        '78496-7',  # Critical care medicine Consult note
        '78498-3',  # Cardiopulmonary Consult note
        '78567-5',  # Pain medicine Consult note
        '78568-3',  # Palliative care Consult note
        '78726-7',  # Pediatrics Consult note
        '78732-5',  # Trauma Consult note
        '78738-2',  # Sports medicine Consult note
        '79428-9',  # Rapid response team Consult note
        '80396-5',  # Gynecologic oncology Consult note
        '80575-4',  # Cardiac surgery Consult note
        '80664-6',  # Clinical neurophysiology Consult note
        '80666-1',  # Vascular neurology Consult note
        '80673-7',  # Maternal and fetal medicine Consult note
        '80736-2',  # Blood banking and transfusion medicine Consult note
        '80801-4',  # Surgical oncology Consult note
        '81191-9',  # Chemical pathology Consult note
        '81192-7',  # Clinical pathology Consult note
        '81193-5',  # Medical microbiology - pathology Consult note
        '81196-8',  # Clinical pharmacology Consult note
        '82356-7',  # Obstetrics Midwife Consult note
        '82359-1',  # Reproductive endocrinology and infertility Consult note
        '83570-2',  # Diabetology Nurse Consult note
        '83578-5',  # Hematology+Medical oncology Nurse Consult note
        '83609-8',  # Psychiatry Nurse Consult note
        '83621-3',  # Geriatric medicine Nurse Consult note
        '83653-6',  # Pain medicine Team Consult note
        '83685-8',  # Hematology+Medical oncology Attending Consult note
        '83720-3',  # Physical medicine and rehab Long term care facility Consult note
        '83722-9',  # Physical medicine and rehabilitation Nurse Consult note
        '83868-0',  # Preventive medicine Consult note
        '83873-0',  # Primary care Consult note
        '83888-8',  # Psychiatry Long term care facility Consult note
        '83909-2',  # Public health Consult note
        '83912-6',  # Pulmonary Patient's home Consult note
        '83926-6',  # Recreational therapy Long term care facility Consult note
        '83931-6',  # Research Consult note
        '83941-5',  # Respiratory therapy Patient's home Consult note
        '83960-5',  # Social worker Patient's home Consult note
        '83967-0',  # Social worker Long term care facility Consult note
        '83984-5',  # Speech-language pathology Consult note
        '83992-8',  # Spinal cord injury medicine Consult note
        '83996-9',  # Spinal cord injury medicine Patient's home Consult note
        '84035-5',  # Transplant surgery Consult note
        '84071-0',  # Chiropractic medicine Consult note
        '84115-5',  # Geriatric medicine Long term care facility Consult note
        '84126-2',  # Surgery of the hand Consult note
        '84131-2',  # Primary care Patient's home Consult note
        '84142-9',  # Hematology+Medical oncology Hospital Consult note
        '84145-2',  # Hematology+Medical oncology Outpatient Consult note
        '84152-8',  # Hepatology Consult note
        '84173-4',  # Interventional cardiology Consult note
        '84190-8',  # Medical toxicology Consult note
        '84213-8',  # Mental health Team Consult note
        '84231-0',  # Neurology Telehealth Consult note
        '84241-9',  # Nurse Consult note
        '84280-7',  # Nutrition and dietetics Patient's home Consult note
        '84292-2',  # Nutrition and dietetics Team Consult note
        '84303-7',  # Wound care management Consult note
        '84312-8',  # Vocational rehabilitation Consult note
        '84324-3',  # Physical therapy Long term care facility Consult note
        '84349-0',  # Pastoral care Consult note
        '84352-4',  # Palliative care Team Consult note
        '84358-1',  # Palliative care Patient's home Consult note
        '84394-6',  # Occupational therapy Patient's home Consult note
        '84398-7',  # Occupational therapy Long term care facility Consult note
        '85174-1',  # Custodial care facility Consult note
        '85208-7',  # Telehealth Consult note
        '85222-8',  # Case manager Consult note
        '85232-7',  # Tumor board Consult note
        '85237-6',  # Acupuncture Consult note
        '85238-4',  # Internal medicine Consult note
        '85517-1',  # Bariatric surgery Consult note
        '85519-7',  # Child and adolescent psychology Consult note
        '85866-2',  # Sleep medicine Consult note
        '85871-2',  # Womens health Consult note
        '85882-9',  # Ophthalmology Teleimaging Consult note
        '85884-5',  # Epilepsy Consult note
        '85886-0',  # Clinical cardiac electrophysiology Consult note
        '85890-2',  # Neuropsychology Consult note
        '85899-3',  # Community health care Consult note
        '86451-2',  # Ethics Consult note
        '87233-3',  # Dialysis Consult note
        '87254-9',  # Addiction medicine Consult note
        '87627-6',  # Brain injury Consult note
        '88351-2',  # Polytrauma Consult note
        '88640-8',  # Community health care Adult day care center Consult note
        '88644-0',  # Outpatient hospital Consult note
        '89031-9',  # Hematology Telehealth Consult note
        '89032-7',  # Nephrology Telehealth Consult note
        '89033-5',  # Urology Telehealth Consult note
        '89216-6',  # Gynecology Consult note
        '89227-3',  # Obstetrics Consult note
        '89446-9',  # Obstetrics Hospital Consult note
        '89447-7',  # Gynecology Hospital Consult note
        '89551-6',  # Environmental health Consult note
        '90006-8',  # Pharmacogenomics Consult note
        '90012-6',  # Burn management Consult note
        '90343-5',  # Nuclear medicine Consult note
        '90354-2',  # Obesity medicine Consult note
        '90709-7',  # Adolescent medicine Consult note
        '90710-5',  # Developmental-behavioral pediatrics Consult note
        '90712-1',  # Sleep medicine Telehealth Consult note
        '90714-7',  # Pain medicine Telehealth Consult note
        '90715-4',  # Palliative care Telehealth Consult note
        '90717-0',  # Heart failure+Transplant cardiology Consult note
        '90771-7',  # Heart failure Consult note
        '91986-0',  # Integrative medicine Consult note
        '92910-9',  # Spinal surgery Consult note
        '92912-5',  # Eating disorders Consult note
        '92915-8',  # Immunology Consult note
        '93024-8',  # Pharmacist Consult note
        '93413-3',  # Mechanical circulatory support Consultation note
        '93955-3',  # Wound, Ostomy, and Continence Care Consult note
    }
    SNOMEDCT = {
        '371530004',  # Clinical consultation report (record artifact)
        '371531000',  # Report of clinical encounter (record artifact)
        '371545006',  # Confirmatory consultation report (record artifact)
    }




# type: ignore
from typing import cast

from canvas_workflow_kit import events, settings
from canvas_workflow_kit.protocol import (
    STATUS_DUE,
    STATUS_SATISFIED,
    ClinicalQualityMeasure,
    ExternallyAwareClinicalQualityMeasure,
    ProtocolResult
)


class CustomerScanHCCSuspectAlert(ExternallyAwareClinicalQualityMeasure, ClinicalQualityMeasure):
    """
    SCAN specific protocol that displays Suspect HCC data, if the patient has any.
    There is currently no action to take, so no recommendations are made.
    """

    class Meta:
        title = 'SCAN: HCC Suspect Alert'

        version = '2020-09-04v1'

        description = (
            'SCAN specific protocol that displays Suspect HCC data, if the patient has any. '
            'There is currently no action to take, so no recommendations are made.')
        information = 'https://docs.google.com/document/d/1NW4b3OT_yPeZy_oIKt_DCCeQ0egDaWFE/edit'

        identifiers = ['CustomerScanHCCSuspectAlertV1']

        types = ['CQM']

        responds_to_event_types = [
            events.HEALTH_MAINTENANCE,
        ]

        authors = [
            'SCAN',
        ]

        compute_on_change_types = [
            ClinicalQualityMeasure.CHANGE_PROTOCOL_OVERRIDE,
            ClinicalQualityMeasure.CHANGE_SUSPECT_HCC,
        ]

        funding_source = 'SCAN'

        references = ['Written by SCAN']

    @classmethod
    def enabled(cls) -> bool:
        return cast(bool, settings.IS_HEALTHCHEC)

    def in_initial_population(self) -> bool:
        return True

    def in_denominator(self) -> bool:
        """
        Denominator: Equals Initial Population

        Exclusions: None

        Exceptions: None
        """

        return self.in_initial_population()

    def in_numerator(self) -> bool:
        """
        Patient is in the numerator if they have no suspect hccs.
        """

        return len(self.patient.suspect_hccs) == 0

    def craft_satisfied_result(self):
        """
        Satisfied if in numerator. (See numerator description)
        """
        result = ProtocolResult()

        result.due_in = -1
        result.status = STATUS_SATISFIED

        result.add_narrative(f'{self.patient.first_name} does not have any Suspect HCC data.')

        return result

    def craft_unsatisfied_result(self):
        """
        Unsatisfied if not in the numerator.
        """
        result = ProtocolResult()

        result.due_in = -1
        result.status = STATUS_DUE

        previous_conditions = []
        hccs_implied_by_clinical_data = []

        for suspect_hcc in self.patient.suspect_hccs:
            if suspect_hcc['kind'] == 'previous_condition':
                previous_conditions.append(suspect_hcc)
            elif suspect_hcc['kind'] == 'implied_by_clinical_data':
                hccs_implied_by_clinical_data.append(suspect_hcc)

        if previous_conditions:
            previous_conditions_narrative = ('─────────────────────\n'
                                             'Previous conditions for consideration:\n')
            for suspect_hcc in previous_conditions:
                previous_conditions_narrative += f"    {suspect_hcc['previousCondition']}\n"
            result.add_narrative(previous_conditions_narrative)

        if hccs_implied_by_clinical_data:
            hccs_narrative = ('─────────────────────\nAI suspects for consideration:\n')
            for suspect_hcc in hccs_implied_by_clinical_data:
                hccs_narrative += f"\n    {suspect_hcc['hccDescription']}\n"
                for reason in suspect_hcc['reasons']:
                    hccs_narrative += f'        • {reason}\n'
            result.add_narrative(hccs_narrative)

        return result



from ..value_set import ValueSet


class StatinAllergen(ValueSet):
    """
    **Clinical Focus:** The purpose of this value set is to represent concepts of an allergy to statin medications.

    **Data Element Scope:** This value set may use a model element related to Allergy/Intolerance.

    **Inclusion Criteria:** Includes concepts that define an allergy to statin medications.

    **Exclusion Criteria:** No exclusions.

    ** Used in:** CMS347v5
    """

    VALUE_SET_NAME = 'Statin Allergen'
    OID = '2.16.840.1.113762.1.4.1110.42'
    DEFINITION_VERSION = '20210220'
    EXPANSION_VERSION = 'eCQM Update 2021-05-06'

    RXNORM = {
        '6472',  # lovastatin
        '36567',  # simvastatin
        '41127',  # fluvastatin
        '42463',  # pravastatin
        '83367',  # atorvastatin
        '301542',  # rosuvastatin
        '861634',  # pitavastatin
    }
    SNOMEDCT = {
        '372912004',  # Substance with 3-hydroxy-3-methylglutaryl-coenzyme A reductase inhibitor mechanism of action (substance)
    }




# type: ignore
from typing import Dict, List, Optional, cast

import arrow

from cached_property import cached_property

from canvas_workflow_kit import events, settings
from canvas_workflow_kit.protocol import (
    STATUS_DUE,
    STATUS_SATISFIED,
    ClinicalQualityMeasure,
    CodingStruct,
    ExternallyAwareClinicalQualityMeasure,
    ProtocolResult
)
from canvas_workflow_kit.recommendation import StructuredAssessmentRecommendation
from canvas_workflow_kit.timeframe import Timeframe
from canvas_workflow_kit.value_set.customer_value_sets.scan import (
    QuestionnaireComprehensiveGeriatricAssessment,
    StructuredAssessmentQuestionnaireCognitiveDisorders
)


class CustomerScanDementiaAssessment(ExternallyAwareClinicalQualityMeasure,
                                     ClinicalQualityMeasure):
    """
    A clinical quality measure (which is a type of protocol) for the customer SCAN.
    More information at https://docs.google.com/document/d/1NW4b3OT_yPeZy_oIKt_DCCeQ0egDaWFE/edit
    """

    class Meta:
        title = 'SCAN: Dementia Assessment'
        version = '2020-10-13v1'
        description = 'A clinical quality measure (which is a type of protocol) for the customer SCAN.'  # noqa: E501
        information = 'https://docs.google.com/document/d/1NW4b3OT_yPeZy_oIKt_DCCeQ0egDaWFE/edit'

        identifiers = ['CustomerScanDementiaAssessmentV1']

        types = ['CQM']

        responds_to_event_types = [
            events.HEALTH_MAINTENANCE,
        ]

        authors = [
            'SCAN',
        ]
        references = ['Written by SCAN']

        compute_on_change_types = [
            ClinicalQualityMeasure.CHANGE_PROTOCOL_OVERRIDE,
            ClinicalQualityMeasure.CHANGE_INTERVIEW,
        ]

        funding_source = 'SCAN'

    @classmethod
    def enabled(cls) -> bool:
        return cast(bool, settings.IS_HEALTHCHEC)

    def in_initial_population(self) -> bool:
        return True

    def interview_lookback_timeframe(self) -> Timeframe:
        return Timeframe(start=self.timeframe.end.shift(days=-30), end=self.timeframe.end)

    @cached_property
    def recent_comprehensive_geriatric_assessments(self) -> List[Dict]:
        interviews = self.patient.interviews.find(
            QuestionnaireComprehensiveGeriatricAssessment).within(
                self.interview_lookback_timeframe())

        for interview in interviews:
            interview['groupedResponses'] = self.group_question_responses(interview)

        return interviews

    @cached_property
    def latest_minicog_response(self) -> Optional[Dict]:
        minicog_question_coding = CodingStruct(
            code='SCAN_MINICOG_QUESTION_CODE', system='INTERNAL')

        return self.latest_response_to_question(question_coding=minicog_question_coding)

    def latest_minicog_response_date(self) -> Optional[arrow.Arrow]:
        if self.latest_minicog_response:
            return arrow.get(self.latest_minicog_response.get('note_timestamp'))
        return None

    def latest_response_to_question(self, question_coding: CodingStruct) -> Optional[Dict]:
        for interview in sorted(
                self.recent_comprehensive_geriatric_assessments,
                key=lambda i: arrow.get(i['noteTimestamp']),
                reverse=True):
            question_responses = interview.get('groupedResponses', {}).get(
                question_coding.code, [])

            if question_responses:
                return {
                    'note_timestamp': interview['noteTimestamp'],
                    'question_coding': question_coding,
                    'responses': question_responses,
                }

    def last_interview(self, questionnaire_valueset, timeframe=None) -> Optional[Dict]:
        interview = self.patient.interviews.find(questionnaire_valueset).within(
            timeframe or self.timeframe).last()

        if not interview:
            return None

        return interview

    @cached_property
    def last_cognitive_disorders_assessment(self) -> Optional[Dict]:
        return self.last_interview(
            questionnaire_valueset=StructuredAssessmentQuestionnaireCognitiveDisorders,
            timeframe=self.interview_lookback_timeframe())

    def last_cognitive_disorders_assessment_date(self) -> Optional[arrow.Arrow]:
        if self.last_cognitive_disorders_assessment:
            return arrow.get(self.last_cognitive_disorders_assessment['noteTimestamp'])
        return None

    def cognitive_disorders_assessed_on_or_after(self, reference_date: arrow.Arrow) -> bool:
        # It can't have been assessed after a date that doesn't exist!
        if not reference_date:
            return False

        # If it hasn't been assessed, it certainly wasn't assessed after the
        # reference date.
        if not self.last_cognitive_disorders_assessment_date():
            return False

        return self.last_cognitive_disorders_assessment_date() >= reference_date

    def in_denominator(self) -> bool:
        """
        Denominator: Equals Initial Population

        Exclusions: Exclude patients who were in hospice care during the measurement year

        Exceptions: None
        """
        if not self.in_initial_population():
            return False

        if self.patient.hospice_within(self.timeframe):
            return False

        return True

    def in_numerator(self) -> bool:
        """
        Represents whether this patient should be counted in the numerator of a fraction
        representing how many patients have been properly appraised regarding the
        CustomerScanDementiaAssessment clinical quality measure.
        A patient have been properly appraised if they either indicated via the assessment that they
        did not need further assessment or, if they indicated that they did, they received that
        assessment.
        """

        low_minicog_score_reported = False

        if self.latest_minicog_response:
            if set(self.latest_minicog_response['responses']).intersection(
                    set(['minicog_score_0', 'minicog_score_1', 'minicog_score_2'])):
                low_minicog_score_reported = True

        no_low_minicog_score_reported = not low_minicog_score_reported

        cognitive_disorders_assessed = self.cognitive_disorders_assessed_on_or_after(
            self.latest_minicog_response_date())

        return no_low_minicog_score_reported or (low_minicog_score_reported and
                                                 cognitive_disorders_assessed)

    def craft_satisfied_result(self):
        """
        Creates and returns a STATUS_SATISFIED-status protocol result with a narrative explaining
        the status of a recent CustomerScanDementiaAssessment if there was one.
        """
        result = ProtocolResult()

        result.due_in = -1
        result.status = STATUS_SATISFIED

        cognitive_disorders_assessed = self.cognitive_disorders_assessed_on_or_after(
            self.latest_minicog_response_date())

        if cognitive_disorders_assessed:
            # Assessment was administered
            result.add_narrative(
                f'{self.patient.first_name} had a cognitive disorders assessment '
                f'{self.display_date(self.last_cognitive_disorders_assessment_date())}.')
        else:
            if self.latest_minicog_response_date():
                result.add_narrative(
                    f"{self.patient.first_name}'s most recent minicog score was not less than 3.")
            else:
                result.add_narrative(f'{self.patient.first_name} has not had a '
                                     'low minicog score in the past 30 days.')

        return result

    def craft_unsatisfied_result(self):
        """
        Creates and returns a STATUS_DUE-status protocol result with a recommendation for a
        CUSTOMER_SCAN_RECOMMEND_COGNITIVE_DISORDERS_ASSESSMENT.
        """
        result = ProtocolResult()

        result.due_in = -1
        result.status = STATUS_DUE

        result.add_narrative(f'{self.patient.first_name} had a minicog score < 3 '
                             f'{self.display_date(self.latest_minicog_response_date())}.')

        result.add_recommendation(
            StructuredAssessmentRecommendation(
                key='CUSTOMER_SCAN_RECOMMEND_COGNITIVE_DISORDERS_ASSESSMENT',
                rank=1,
                button='Assess',
                patient=self.patient,
                questionnaires=[StructuredAssessmentQuestionnaireCognitiveDisorders],
                title='Assess Cognitive Disorder'))

        return result



from protocols import events
from protocols.protocol import ClinicalQualityMeasure
from protocols.value_set.v2021 import (
    BehavioralneuropsychAssessment, CareServicesInLongTermResidentialFacility, CognitiveAssessment,
    DementiaMentalDegenerations, Ethnicity, HomeHealthcareServices, NursingFacilityVisit,
    OccupationalTherapyEvaluation, OfficeVisit, OncAdministrativeSex, OutpatientConsultation,
    PatientProviderInteraction, Payer, PsychVisitDiagnosticEvaluation, PsychVisitPsychotherapy,
    Race, StandardizedToolsForAssessmentOfCognition)


class ClinicalQualityMeasure149v9(ClinicalQualityMeasure):
    """
    Dementia: Cognitive Assessment

    Description: Percentage of patients, regardless of age, with a diagnosis of dementia for whom
    an assessment of cognition is performed and the results reviewed at least once within a
    12-month period

    Definition: Cognition can be assessed by the clinician during the patient's clinical history.
    Cognition can also be assessed by direct examination of the patient using one of a number of
    instruments, including several originally developed and validated for screening purposes. This
    can also include, where appropriate, administration to a knowledgeable informant. Examples
    include, but are not limited to:
    -Blessed Orientation-Memory-Concentration Test (BOMC)
    -Montreal Cognitive Assessment (MoCA)
    -St. Louis University Mental Status Examination (SLUMS)
    -Mini-Mental State Examination (MMSE) [Note: The MMSE has not been well validated for non-
    Alzheimer's dementias]
    -Short Informant Questionnaire on Cognitive Decline in the Elderly (IQCODE)
    -Ascertain Dementia 8 (AD8) Questionnaire
    -Minimum Data Set (MDS) Brief Interview of Mental Status (BIMS) [Note: Validated for use with
    nursing home patients only]
    -Formal neuropsychological evaluation
    -Mini-Cog

    Rationale: An estimated 5.8 million of adults in the US were living with dementia in 2019.
    Dementia is often characterized by the gradual onset and continuing cognitive decline in one or
    more domains including memory, communication and language, ability to focus or pay attention,
    reasoning and judgment and visual perception (Alzheimer’s Association, 2019). Cognitive
    deterioration represents a major source of morbidity and mortality and poses a significant
    burden on affected individuals and their caregivers (Daviglus et al., 2010). Although cognitive
    deterioration follows a different course depending on the type of dementia, significant rates
    of decline have been reported. For example, one study found that the annual rate of decline for
    Alzheimer's disease patients was more than four times that of older adults with no cognitive
    impairment (Wilson et al., 2010). Nevertheless, measurable cognitive abilities remain
    throughout the course of dementia (American Psychiatric Association, 2007). Initial and ongoing
    assessments of cognition are fundamental to the proper management of patients with dementia.
    These assessments serve as the basis for identifying treatment goals, developing a treatment
    plan, monitoring the effects of treatment, and modifying treatment as appropriate.

    Guidance: Use of a standardized tool or instrument to assess cognition other than those listed
    will meet numerator performance. Standardized tools can be mapped to the concept "Intervention,
    Performed": "Cognitive Assessment" included in the numerator logic below.

    The requirement of two or more visits is to establish that the eligible professional or
    eligible clinician has an existing relationship with the patient.

    In recognition of the growing use of integrated and team-based care, the diagnosis of dementia
    and the assessment of cognitive function need not be performed by the same provider or
    clinician.

    The DSM-5 has replaced the term dementia with major neurocognitive disorder and mild
    neurocognitive disorder. For the purposes of this measure, the terms are equivalent.

    This eCQM is a patient-based measure.

    This version of the eCQM uses QDM version 5.5.  Please refer to the eCQI resource center
    (https://ecqi.healthit.gov/qdm) for more information on the QDM.

    More information: https://ecqi.healthit.gov/ecqm/measures/cms149v9
    """

    title = 'Dementia: Cognitive Assessment'

    identifiers = ['CMS149v9']

    types = ['CQM']

    responds_to_event_types = [
        events.HEALTH_MAINTENANCE,
    ]

    authors = [
        'American Medical Association (AMA)',
        'PCPI(R) Foundation (PCPI[R])',
    ]

    references = [
        'Alzheimer’s Association. (2019). 2019 Alzheimer’s Disease Facts and Figures. Alzheimer’s & Dementia, 15(3), 321-387. Retrieved from https://alz.org/media/Documents/alzheimers-facts-and-figures-2019-r.pdf',
        'American Psychiatric Association, Work Group on Alzheimer’s Disease and Other Dementias. (2007). Practice Guideline for the Treatment of Patients with Alzheimer’s Disease and Other Dementias. Retrieved from https://psychiatryonline.org/pb/assets/raw/sitewide/practice_guidelines/guidelines/alzheimers.pdf',
        'American Psychiatric Association. (2016). Practice guideline on the use of antipsychotics to treat agitation or psychosis in patients with dementia. Retrieved from https://psychiatryonline.org/doi/pdf/10.1176/appi.books.9780890426807',
        'California Department of Public Health. (2017). California guidelines for Alzheimer’s disease management, 2017. Retrieved from https://www.cdph.ca.gov/Programs/CCDPHP/DCDIC/CDCB/CDPH%20Document%20Library/Alzheimers\'%20Disease%20Program/ALZ-CareGuidelines.pdf',
        'Daviglus M.L., Bell, C.C., Berrettini, W., Bowen, P.E., Connolly, E.S., Cox, N.J.,…Trevisan, M. (2010). National Institutes of Health State-of-the-Science Conference Statement: Preventing Alzheimer’s Disease and Cognitive Decline. NIH Consensus and State-of-the-Science Statements, 27(4), 1-24. Retrieved from https://consensus.nih.gov/2010/alzstatement.htm',
        'Fazio, S., Pace, D., Maslow, K., Zimmerman, S., & Kallmyer, B. (2018). Alzheimer’s Association Dementia Care Practice Recommendations. The Gerontologist, 58(S1), S1-S9. [Supplemental material]. Retrieved from the Alzheimer’s Association website: https://academic.oup.com/gerontologist/article/58/suppl_1/S1/4816759',
        'U.S. Department of Health and Human Services, Assistant Secretary for Planning and Evaluation, Office of Disability, Aging and Long-Term Care Policy. (2016). Examining models of dementia care: Final report. (ASPE Final Report No. 0212704.017.000.001). Retrieved from https://aspe.hhs.gov/system/files/pdf/257216/ExamDCMod.pdf',
        'Wilson, R. S., Aggarwal, N. T., Barnes, L. L., Mendes de Leon, C.F., Herbert, L.E., & Evans, D.A. (2010). Cognitive decline in incident Alzheimer disease in a community population. Neurology, 74(12), 951-955.',
    ]

    funding_source = ''

    def in_initial_population(self):
        """
        Initial population: All patients, regardless of age, with a diagnosis of dementia
        """
        pass

    def in_denominator(self):
        """
        Denominator: Equals Initial Population

        Exclusions: None

        Exceptions: Documentation of patient reason(s) for not assessing cognition
        """
        return self.in_initial_population()

    def in_numerator(self):
        """
        Numerator: Patients for whom an assessment of cognition is performed and the results
        reviewed at least once within a 12-month period

        Exclusions: Not Applicable
        """
        pass


    def compute_results(self):
        """
        Clinical recommendation: Ongoing assessment includes periodic monitoring of the development
        and evolution of cognitive and noncognitive psychiatric symptoms and their response to
        intervention (Category I). Both cognitive and noncognitive neuropsychiatric and behavioral
        symptoms of dementia tend to evolve over time, so regular monitoring allows detection of
        new symptoms and adaptation of treatment strategies to current needs... Cognitive symptoms
        that almost always require assessment include impairments in memory, executive function,
        language, judgment, and spatial abilities. It is often helpful to track cognitive status
        with a structured simple examination (American Psychiatric Association, 2007).

        The American Psychiatric Association recommends that patients with dementia be assessed for
        the type, frequency, severity, pattern, and timing of symptoms (Category 1C). Quantitative
        measures provide a structured replicable way to document the patient's baseline symptoms
        and determine which symptoms (if any) should be the target of intervention based on factors
        such as frequency of occurrence, magnitude, potential for associated harm to the patient or
        others, and associated distress to the patient. The exact frequency at which measures are
        warranted will depend on clinical circumstances. However, use of quantitative measures as
        treatment proceeds allows more precise tracking of whether nonpharmacological and
        pharmacological treatments are having their intended effect or whether a shift in the
        treatment plan is needed (American Psychiatric Association, 2016).

        Conduct and document an assessment and monitor changes in cognitive status using a reliable
        and valid instrument, e.g., MoCA (Montreal Cognitive Assessment), AD8 (Ascertian Dementia
        8) or other tool. Cognitive status should be reassessed periodically to identify sudden
        changes, as well as to monitor the potential beneficial or harmful effects of environmental
        changes (including safety, care needs, and abuse and/or neglect), specific medications
        (both prescription and non-prescription, for appropriate use and contraindications), or
        other interventions. Proper assessment requires the use of a standardized, objective
        instrument that is relatively easy to use, reliable (with less variability between
        different assessors), and valid (results that would be similar to gold-standard
        evaluations) (California Department of Public Health, 2017).

        Recommendation: Perform regular, comprehensive person-centered assessments and timely
        interim assessments.
        Assessments, conducted at least every 6 months, should prioritize issues that help the
        person with dementia to live fully. These include assessments of the individual and care
        partner’s relationships and subjective experience and assessment of cognition, behavior,
        and function, using reliable and valid tools. Assessment is ongoing and dynamic, combining
        nomothetic (norm based) and idiographic (individualized) approaches (Fazio, Pace, Maslow,
        Zimmerman, & Kallmyer, 2018)

        Recommendation: Assess cognitive status, functional abilities, behavioral and psychological
        symptoms of dementia, medical status, living environment, and safety. Reassess regularly
        and when there is a significant change in condition (U.S. Department of Health and Human
        Services, 2016).
        """
        pass



import arrow

from canvas_workflow_kit import settings

from canvas_workflow_kit.builtin_cqms.hcc004v1_dysrhythmia_suspect import Hcc004v1
from canvas_workflow_kit.patient_recordset import ConditionRecordSet, MedicationRecordSet
from canvas_workflow_kit.tests.base import SDKBaseTest
from canvas_workflow_kit.timeframe import Timeframe


class TestHcc004v1(SDKBaseTest):

    def setUp(self):
        self.mocks_path = f'{settings.BASE_DIR}/builtin_cqms/tests/mock_data'

    def test_enabled(self):
        self.assertTrue(Hcc004v1.enabled())

    def test_description(self):
        expected = ('All patients with potential dysrhythmia based on '
                    'an active medication without associated active problem.')
        self.assertEqual(expected, Hcc004v1._meta.description)

    def test_information(self):
        expected = ('https://canvas-medical.zendesk.com/hc/en-us/articles/360059083773-Dysrhythmia-Suspects-HCC004v1')
        self.assertEqual(expected, Hcc004v1._meta.information)

    def test_change_types(self):
        result = Hcc004v1._meta.compute_on_change_types
        expected = [
            'protocol_override',
            'condition',
            'medication',
        ]
        self.assertEqual(len(expected), len(result))
        for change in expected:
            self.assertIn(change, result)

    def test_in_initial_population(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)
        patient = self.load_patient('hcc004v1')
        tested = Hcc004v1(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_initial_population())

    def test_in_denominator_patient_with_active_antiarrhythmics_medication(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        patient = self.load_patient('hcc004v1')
        patient.medications = MedicationRecordSet([{
            'periods': [{
                'from': '2018-08-18',
                'to': None
            }],
            'clinicalStatus': 'active',
            'coding': [{
                'id': 35,
                'system': 'http://www.fdbhealth.com/',
                'version': '',
                'code': '564460'
            }]
        }])
        tested = Hcc004v1(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_denominator())

        patient = self.load_patient('hcc004v1')
        patient.active_only = True
        patient.medications = MedicationRecordSet([{
            'periods': [{
                'from': '2018-03-08',
                'to': '2018-08-18'
            }],
            'clinicalStatus': 'active',
            'coding': [{
                'id': 35,
                'system': 'http://www.fdbhealth.com/',
                'version': '',
                'code': '564460'
            }]
        }])
        tested = Hcc004v1(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_denominator())

        patient = self.load_patient('hcc004v1')
        patient.active_only = True
        patient.medications = MedicationRecordSet([{
            'periods': [{
                'from': '2018-03-08',
                'to': '2018-08-24'
            }],
            'clinicalStatus': 'active',
            'coding': [{
                'id': 35,
                'system': 'http://www.fdbhealth.com/',
                'version': '',
                'code': '564460'
            }]
        }])
        tested = Hcc004v1(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_denominator())

    def test_in_denominator_patient_with_active_medication_not_antiarrhythmics(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        patient = self.load_patient('hcc004v1')
        patient.medications = MedicationRecordSet([{
            'periods': [{
                'from': '2018-08-18',
                'to': None
            }],
            'clinicalStatus': 'active',
            'coding': [{
                'id': 35,
                'system': 'http://www.fdbhealth.com/',
                'version': '',
                'code': '564017'
            }]
        }])
        tested = Hcc004v1(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_denominator())

    def test_in_denominator_patient_without_active_medication(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)
        patient = self.load_patient('hcc004v1')
        tested = Hcc004v1(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_denominator())

    def test_in_numerator_patient_without_active_condition(self):
        now = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=now, end=now)
        patient = self.load_patient('hcc004v1')
        tested = Hcc004v1(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_numerator())

    def test_in_numerator_patient_with_active_condition_in_dysrhythmia_class(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)

        patient = self.load_patient('hcc004v1')
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'I420'
            }],
            'periods': [{
                'from': '2018-08-23',
                'to': None
            }],
        }])
        tested = Hcc004v1(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_numerator())

        patient = self.load_patient('hcc004v1')
        patient.active_only = True
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'I420'
            }],
            'periods': [{
                'from': '2018-05-23',
                'to': '2018-08-20'
            }],
        }])
        tested = Hcc004v1(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_numerator())

        patient = self.load_patient('hcc004v1')
        patient.active_only = True
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'I420'
            }],
            'periods': [{
                'from': '2018-08-01',
                'to': '2018-08-27'
            }],
        }])
        tested = Hcc004v1(patient=patient, timeframe=timeframe)
        self.assertFalse(tested.in_numerator())

    def test_in_numerator_patient_with_not_active_condition_in_dysrhythmia_class(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)
        patient = self.load_patient('hcc004v1')
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'I420'
            }],
            'periods': [{
                'from': '2016-08-23',
                'to': '2017-08-22'
            }],
        }])
        tested = Hcc004v1(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_numerator())

    def test_in_numerator_patient_with_active_condition_not_in_dysrhythmia_class(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)
        patient = self.load_patient('hcc004v1')
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'I240'
            }],
            'periods': [{
                'from': '2018-08-23',
                'to': None
            }],
        }])
        tested = Hcc004v1(patient=patient, timeframe=timeframe)
        self.assertTrue(tested.in_numerator())

    def test_compute_results_patient_without_active_medication(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)
        patient = self.load_patient('hcc004v1')
        tested = Hcc004v1(patient=patient, timeframe=timeframe)
        result = tested.compute_results()
        self.assertIsNotApplicable(result)

    def test_compute_results_patient_with_active_antiarrhythmics_med_with_dysrhythmia_dx(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)
        patient = self.load_patient('hcc004v1')
        patient.medications = MedicationRecordSet([{
            'periods': [{
                'from': '2018-08-18',
                'to': None
            }],
            'clinicalStatus': 'active',
            'coding': [{
                'id': 35,
                'system': 'http://www.fdbhealth.com/',
                'version': '',
                'code': '564460'
            }]
        }])
        patient.conditions = ConditionRecordSet([{
            'clinicalStatus': 'active',
            'coding': [{
                'system': 'ICD-10',
                'code': 'I420'
            }],
            'periods': [{
                'from': '2018-08-23',
                'to': None
            }],
        }])
        tested = Hcc004v1(patient=patient, timeframe=timeframe)
        result = tested.compute_results()
        self.assertEqual('satisfied', result.status)
        text = ''
        self.assertEqual(text, result.narrative)
        self.assertEqual([], result.recommendations)
        self.assertIsNone(result.due_in)

    def test_compute_results_patient_with_active_antiarrhythmics_med_without_dysrhythmia_dx(self):
        start = arrow.get('2017-08-23 13:24:56')
        end = arrow.get('2018-08-23 13:24:56')
        timeframe = Timeframe(start=start, end=end)
        patient = self.load_patient('hcc004v1')
        patient.medications = MedicationRecordSet([{
            'periods': [{
                'from': '2018-08-18',
                'to': None
            }],
            'clinicalStatus': 'active',
            'coding': [{
                'id': 35,
                'system': 'http://www.fdbhealth.com/',
                'version': '',
                'code': '564460'
            }]
        }])
        tested = Hcc004v1(patient=patient, timeframe=timeframe)
        result = tested.compute_results()
        self.assertEqual('due', result.status)
        text = (
            'Jim has an active medication on the Medication List commonly used for Dysrhythmia. '
            'There is no associated condition on the Conditions List.')
        self.assertEqual(text, result.narrative)
        expected = [
            {
                'title': ('Consider updating the Conditions List to include Dysrhythmia '
                          'related problem as clinically appropriate.'),
                'command': {
                    'key': 'diagnose'
                },
                'key': 'HCC004v1_RECOMMEND_DIAGNOSE_DYSRHYTHMIA',
                'rank': 1,
                'button': 'Diagnose',
            },
        ]  # yapf: disable
        self.helper_exam_recommended(expected, result.recommendations)
        self.assertEqual(-1, result.due_in)

    def helper_exam_recommended(self, expected, result):
        for exp, recommendation in zip(expected, result):
            self.assertEqual(exp['key'], recommendation.key)
            self.assertEqual(exp['rank'], recommendation.rank)
            self.assertEqual(exp['button'], recommendation.button)
            self.assertEqual(exp['title'], recommendation.title)
            self.assertEqual(exp['command']['key'], recommendation.command['key'])



# type: ignore
from typing import List

import arrow

from cached_property import cached_property

from canvas_workflow_kit import events
from canvas_workflow_kit.patient_recordset import InterviewRecordSet
from canvas_workflow_kit.protocol import (
    STATUS_DUE,
    STATUS_SATISFIED,
    ClinicalQualityMeasure,
    ProtocolResult
)
from canvas_workflow_kit.recommendation import Recommendation
from canvas_workflow_kit.value_set.specials import (
    Covid19QuestionnaireHighRiskOutreach,
    Covid19QuestionnaireSymptomaticSurveillance
)
from canvas_workflow_kit.value_set.v2017 import (
    ChronicObstructivePulmonaryDisease,
    ImmunocompromisedConditions,
    PersistentAsthma
)
from canvas_workflow_kit.value_set.v2019 import IschemicVascularDisease
from canvas_workflow_kit.value_set.v2020 import (
    ChronicLiverDisease,
    Diabetes,
    KidneyFailure,
    MorbidObesity
)

MINIMUM_AGE = 65


class Ccp002v1(ClinicalQualityMeasure):

    class Meta:
        title = 'COVID-19 High Risk Outreach'

        version = '2020-03-24v1'

        description = f'All patients with {MINIMUM_AGE}+ with chronic conditions to be reached.'
        information = 'https://canvas-medical.zendesk.com/hc/en-us/articles/360059084173-COVID-19-Risk-Assessment-Follow-Up-Protocol'

        identifiers = ['CCP002v1']

        types = ['CCP']

        responds_to_event_types = [
            events.HEALTH_MAINTENANCE,
        ]
        authors = [
            'Canvas Medical Team',
        ]
        show_in_chart = False

        references = [
            'Canvas Medical CCP, https://canvas-medical.zendesk.com/hc/en-us/articles/360059084173-COVID-19-Risk-Assessment-Follow-Up-Protocol'
        ]

        compute_on_change_types = [
            ClinicalQualityMeasure.CHANGE_CONDITION,
            ClinicalQualityMeasure.CHANGE_INTERVIEW,
            ClinicalQualityMeasure.CHANGE_PATIENT,
        ]

    @cached_property
    def date_of_first_questionnaire(self) -> str:
        questionnaire = self.patient.interviews.find(Covid19QuestionnaireSymptomaticSurveillance |
                                                     Covid19QuestionnaireHighRiskOutreach).first()
        if questionnaire:
            return questionnaire[InterviewRecordSet.DATE_FIELD]
        return ''

    @cached_property
    def high_risk_conditions(self) -> List:
        value_sets = [
            ChronicLiverDisease,
            ChronicObstructivePulmonaryDisease,
            Diabetes,
            ImmunocompromisedConditions,
            IschemicVascularDisease,
            KidneyFailure,
            MorbidObesity,
            PersistentAsthma,
        ]
        result: List = []
        for value_set in value_sets:
            result.extend([
                record for record in self.patient.conditions.find(value_set)
                if record['clinicalStatus'] == 'active'
            ])
        return result

    def in_initial_population(self) -> bool:
        """
        Patients MINIMUM_AGE+ with more than two chronic conditions
        """
        return self.patient.age >= MINIMUM_AGE or self.high_risk_conditions

    def in_denominator(self) -> bool:
        """
        Patients in the initial population
        """
        return self.in_initial_population()

    def in_numerator(self) -> bool:
        """
        Patients that have taken one of the COVID questionnaires
        """
        return bool(self.date_of_first_questionnaire)

    def compute_results(self) -> ProtocolResult:
        result = ProtocolResult()
        if self.in_denominator():
            if self.in_numerator():
                result.due_in = -1
                result.status = STATUS_SATISFIED
                date = arrow.get(self.date_of_first_questionnaire).format('ddd, MMM Do YYYY')
                result.add_narrative(
                    f'{self.patient.first_name} took a COVID-19 outreach questionnaire on {date}.')
            else:
                result.add_narrative(
                    f'{self.patient.first_name} should take a COVID-19 outreach questionnaire.')

                if self.patient.age >= MINIMUM_AGE:
                    result.add_narrative(f'{self.patient.first_name} is {MINIMUM_AGE}+')

                if self.high_risk_conditions:
                    has_display = all(
                        [bool('display' in c['coding'][0]) for c in self.high_risk_conditions])
                    if has_display:
                        result.add_narrative(
                            f'{self.patient.first_name} has high risk conditions:')
                        for condition in self.high_risk_conditions:
                            result.add_narrative(condition['coding'][0]['display'])
                    else:
                        result.add_narrative(
                            f'{self.patient.first_name} has {len(self.high_risk_conditions)} high risk conditions'  # noqa: E501
                        )

                result.due_in = 0
                result.status = STATUS_DUE

                narrative = (f'{self.patient.first_name} should be contacted and have a '
                             'COVID-19 questionnaire administered.')
                result.add_recommendation(
                    Recommendation(
                        key='CCP002v1_RECOMMEND_QUESTIONNAIRE',
                        rank=1,
                        button='Communication',
                        title='Complete COVID-19 outreach questionnaire',
                        narrative=narrative,
                        command={'key': 'schedule'}))

        return result




from decouple import Config, config, RepositoryIni

from pathlib import Path

from typing import Any, cast, Dict, Optional

BASE_DIR = Path(__file__).resolve().parent

CONFIG_PATH = Path.home() / '.canvas' / 'config.ini'
DEFAULT_CONFIG_SECTION = 'canvas_cli'


INTEGRATION_SERVICE_ENDPOINT = config('INTEGRATION_SERVICE_ENDPOINT', '')

INTEGRATION_SERVICE_USERNAME = config('INTEGRATION_SERVICE_USERNAME', '')
INTEGRATION_SERVICE_PASSWORD = config('INTEGRATION_SERVICE_PASSWORD', '')

IS_HEALTHCHEC = True


def read_config(config_name: Optional[str] = None) -> Dict[str, Any]:

    if config_name is None:
        config_name = DEFAULT_CONFIG_SECTION

    try:
        ini = RepositoryIni(CONFIG_PATH)
    except FileNotFoundError:
        raise FileNotFoundError(
            f'Please add your configuration at "{CONFIG_PATH}"; you can set '
            'defaults using `canvas-cli create-default-settings`.')

    if not ini.parser.has_section(config_name):
        raise ValueError(
            f'Settings file "{settings.CONFIG_PATH}" does not contain section "{config_name}"; '
            'you can set defaults using `canvas-cli create-default-settings`.')

    ini.SECTION = config_name
    config = Config(ini)

    config: Dict[str, Any] = {
        'url': config('url', cast=str),
        'api_key': config('api-key', cast=str)
    }

    return config



from typing import Dict, List, Optional

import arrow

from canvas_workflow_kit import events
from canvas_workflow_kit.protocol import (
    CONTEXT_REPORT,
    STATUS_DUE,
    STATUS_SATISFIED,
    ClinicalQualityMeasure,
    ExternallyAwareClinicalQualityMeasure,
    ProtocolResult
)
from canvas_workflow_kit.recommendation import (
    ImagingRecommendation,
    LabRecommendation,
    ReferRecommendation
)
from canvas_workflow_kit.timeframe import Timeframe
# @canvas-adr-0006
from canvas_workflow_kit.value_set.specials import CMS130v6CtColonography
# flake8: noqa
from canvas_workflow_kit.value_set.v2018 import (
    AnnualWellnessVisit,
    Colonoscopy,
    CtColonography,
    DischargedToHealthCareFacilityForHospiceCare,
    DischargedToHomeForHospiceCare,
    EncounterInpatient,
    Ethnicity,
    FaceToFaceInteraction,
    FecalOccultBloodTestFobt,
    FitDna,
    FlexibleSigmoidoscopy,
    HomeHealthcareServices,
    HospiceCareAmbulatory,
    MalignantNeoplasmOfColon,
    OfficeVisit,
    OncAdministrativeSex,
    Payer,
    PreventiveCareServicesEstablishedOfficeVisit18AndUp,
    PreventiveCareServicesInitialOfficeVisit18AndUp,
    Race,
    TotalColectomy
)


class ClinicalQualityMeasure130v6(ExternallyAwareClinicalQualityMeasure, ClinicalQualityMeasure):
    """
    Colorectal Cancer Screening

    Description: Percentage of adults 50-75 years of age who had appropriate screening for
    colorectal cancer

    Definition: None

    Rationale: An estimated 132,700 men and women were diagnosed with colon or rectal cancer in
    2015. In the same year, 49,700 were estimated to have died from the disease, making colorectal
    cancer the third leading cause of cancer death in the United States (National Cancer Institute
    2015, American Cancer Society 2015).

    Screening for colorectal cancer is extremely important as there are no signs or symptoms of the
    cancer in the early stages. If the disease is caught in its earliest stages, it has a five-year
    survival rate of 90%; however, the disease is often not caught this early. While screening is
    extremely effective in detecting colorectal cancer, it remains underutilized (American Cancer
    Society 2015).

    The U.S. Preventive Services Task Force has identified fecal occult blood tests, colonoscopy,
    and flexible sigmoidoscopy as effective screening methods (United States Preventive Services
    Task Force 2008).

    Guidance: Patient self-report for procedures as well as diagnostic studies should be recorded
    in "Procedure, Performed" template or "Diagnostic Study, Performed" template in QRDA-1. Patient
    self-report is not allowed for laboratory tests.

    More information: https://ecqi.healthit.gov/sites/default/files/ecqm/measures/CMS130v6.html
    """

    class Meta:
        title = 'Colorectal Cancer Screening'
        version = '2020-02-24v1'
        default_display_interval_in_days = 365 * 10

        description = (
            'Adults 50-75 years of age who have not had appropriate screening for colorectal cancer.'
        )
        information = 'https://ecqi.healthit.gov/sites/default/files/ecqm/measures/CMS130v6.html'

        identifiers = ['CMS130v6']

        types = ['CQM']

        responds_to_event_types = [
            events.HEALTH_MAINTENANCE,
        ]

        authors = [
            'National Committee for Quality Assurance',
        ]

        references = [
            'American Cancer Society. 2015. "Cancer Prevention & Early Detection Facts & Figures 2015-2016." Atlanta: American Cancer Society.',
            'National Cancer Institute. 2015. "SEER Stat Fact Sheets: Colon and Rectum Cancer." Bethesda, MD, http://seer.cancer.gov/statfacts/html/colorect.html',
            'U.S. Preventive Services Task Force (USPSTF). 2008. "Screening for colorectal cancer: U.S. Preventive Services Task Force recommendation statement." Ann Intern Med 149(9):627-37.',
        ]

        compute_on_change_types = [
            ClinicalQualityMeasure.CHANGE_PROTOCOL_OVERRIDE,
            ClinicalQualityMeasure.CHANGE_CONDITION,
            ClinicalQualityMeasure.CHANGE_IMAGING_REPORT,
            ClinicalQualityMeasure.CHANGE_LAB_REPORT,
            ClinicalQualityMeasure.CHANGE_PATIENT,
            ClinicalQualityMeasure.CHANGE_REFERRAL_REPORT,
        ]

    AGE_RANGE_START = 50
    AGE_RANGE_END = 75
    _last_exam = None

    def first_due_in(self) -> Optional[int]:
        if self.patient.age_at(
                self.timeframe.end) < self.AGE_RANGE_START and not self.had_colon_exclusion():
            return (
                self.patient.birthday.shift(years=self.AGE_RANGE_START) - self.timeframe.end).days
        return None

    def had_colon_exclusion(self) -> bool:
        # Past Surgical History: Total abdominal colectomy with proctectomy and ileostomy (procedure)
        if self.patient.conditions.find(TotalColectomy | MalignantNeoplasmOfColon).intersects(
                self.timeframe, still_active=self.patient.active_only):
            return True
        return False

    def in_initial_population(self) -> bool:
        """
        Initial population: Patients 50-75 years of age with a visit during the measurement period
        """
        return (self.patient.age_at_between(self.timeframe.end,
                                            self.AGE_RANGE_START,
                                            self.AGE_RANGE_END) and
                (self.patient.has_visit_within(self.timeframe,
                                               (OfficeVisit |
                                                PreventiveCareServicesEstablishedOfficeVisit18AndUp |
                                                PreventiveCareServicesInitialOfficeVisit18AndUp |
                                                HomeHealthcareServices |
                                                AnnualWellnessVisit))
                 if self.context == CONTEXT_REPORT else True))  # yapf: disable

    def in_denominator(self) -> bool:
        """
        Denominator: Equals Initial Population

        Exclusions: Patients with a diagnosis or past history of total colectomy or colorectal
        cancer.

        Exclude patients who were in hospice care during the measurement year.

        Exceptions: None
        """
        if not self.in_initial_population():
            return False

        if self.patient.hospice_within(self.timeframe):
            return False

        if self.had_colon_exclusion():
            return False

        return True

    def in_numerator(self) -> bool:
        """
        Numerator: Patients with one or more screenings for colorectal cancer. Appropriate
        screenings are defined by any one of the following criteria:
        - Fecal occult blood test (FOBT) during the measurement period
        - Flexible sigmoidoscopy during the measurement period or the four years prior to the
        measurement period
        - Colonoscopy during the measurement period or the nine years prior to the measurement
        period
        - FIT-DNA during the measurement period or the two years prior to the measurement period
        - CT Colonography during the measurement period or the four years prior to the measurement
        period

        Exclusions: Not Applicable
        """
        self._last_exam = None
        period = Timeframe(start=self.timeframe.start, end=self.timeframe.end)
        record = self.patient.lab_reports.find(FecalOccultBloodTestFobt).within(period).last()
        if record:
            self._last_exam = {
                'date': record['originalDate'],
                'what': 'FOBT',
                'days': period.duration,
            }
            return True

        if not self.period_adjustment:
            period.start = self.timeframe.end.shift(years=-3)
        record = self.patient.lab_reports.find(FitDna).within(period).last()
        if record:
            self._last_exam = {
                'date': record['originalDate'],
                'what': 'FIT-DNA',
                'days': period.duration,
            }
            return True

        if not self.period_adjustment:
            period.start = self.timeframe.end.shift(years=-5)
        record = (self.patient.referral_reports.find(FlexibleSigmoidoscopy).within(period).last()
                  or
                  self.patient.imaging_reports.find(FlexibleSigmoidoscopy).within(period).last())
        if record:
            self._last_exam = {
                'date': record['originalDate'],
                'what': 'Flexible sigmoidoscopy',
                'days': period.duration,
            }
            return True

        if not self.period_adjustment:
            period.start = self.timeframe.end.shift(years=-5)
        record = (self.patient.referral_reports.find(CtColonography |
                                                     CMS130v6CtColonography).within(period).last()
                  or
                  self.patient.imaging_reports.find(CtColonography |
                                                    CMS130v6CtColonography).within(period).last())
        if record:
            self._last_exam = {
                'date': record['originalDate'],
                'what': 'CT Colonography',
                'days': period.duration,
            }
            return True

        if not self.period_adjustment:
            period.start = self.timeframe.end.shift(years=-10)
        record = (self.patient.referral_reports.find(Colonoscopy).within(period).last() or
                  self.patient.imaging_reports.find(Colonoscopy).within(period).last())
        if record:
            self._last_exam = {
                'date': record['originalDate'],
                'what': 'Colonoscopy',
                'days': period.duration,
            }
            return True

        return False

    def recent_exam_context(self) -> str:
        recent_related_exam: Dict = {}

        record = (self.patient.referral_reports.find(Colonoscopy).last() or
                  self.patient.imaging_reports.find(Colonoscopy).last())
        if record:
            recent_related_exam = {
                'date': record['originalDate'],
                'what': 'Colonoscopy',
            }

        if not recent_related_exam:
            record = (
                self.patient.referral_reports.find(CtColonography | CMS130v6CtColonography).last()
                or
                self.patient.imaging_reports.find(CtColonography | CMS130v6CtColonography).last())
            if record:
                recent_related_exam = {
                    'date': record['originalDate'],
                    'what': 'CT Colonography',
                }

        if not recent_related_exam:
            record = (self.patient.referral_reports.find(FlexibleSigmoidoscopy).last() or
                      self.patient.imaging_reports.find(FlexibleSigmoidoscopy).last())
            if record:
                recent_related_exam = {
                    'date': record['originalDate'],
                    'what': 'Flexible sigmoidoscopy',
                }

        if not recent_related_exam:
            record = self.patient.lab_reports.find(FitDna).last()
            if record:
                recent_related_exam = {
                    'date': record['originalDate'],
                    'what': 'FIT-DNA',
                }

        if not recent_related_exam:
            record = self.patient.lab_reports.find(FecalOccultBloodTestFobt).last()
            if record:
                recent_related_exam = {
                    'date': record['originalDate'],
                    'what': 'FOBT',
                }

        if recent_related_exam:
            last_date = arrow.get(recent_related_exam['date'])
            return 'Last {what} done {date}.'.format(
                what=recent_related_exam['what'], date=self.display_date(last_date))
        else:
            return 'No relevant exams found.'

    def craft_satisfied_result(self):
        result = ProtocolResult()

        last_date = arrow.get(self._last_exam['date'])

        result.due_in = (last_date.shift(days=self._last_exam['days']) - self.now).days
        result.status = STATUS_SATISFIED
        result.add_narrative('{name} had a {what} {date}.'.format(
            name=self.patient.first_name,
            what=self._last_exam['what'],
            date=self.display_date(last_date)))

        return result

    def craft_unsatisfied_result(self):
        """
        Clinical recommendation: The United States Preventive Services Task Force (2008):

        [1] The USPSTF recommends screening for colorectal cancer using fecal occult blood testing,
        sigmoidoscopy, or colonoscopy in adults, beginning at age 50 years and continuing until age
        75 years (A recommendation).
        [2] The USPSTF concludes that the evidence is insufficient to assess the benefits and harms
        of computed tomographic (CT) colonography and fecal DNA testing as screening modalities for
        colorectal cancer (I statement).
        """
        result = ProtocolResult()

        result.due_in = -1
        result.status = STATUS_DUE

        result.add_narrative('{name} is due for a Colorectal Cancer Screening.'.format(
            name=self.patient.first_name))
        result.add_narrative(self.recent_exam_context())
        result.add_narrative(self.screening_interval_context())

        context = {
            'conditions': [[{
                'code': 'Z1211',
                'system': 'ICD-10',
                'display': 'Encounter for screening for malignant neoplasm of colon',
            }]]
        }

        result.add_recommendation(
            LabRecommendation(
                key='CMS130v6_RECOMMEND_FOBT',
                rank=1,
                button='Order',
                patient=self.patient,
                condition=MalignantNeoplasmOfColon,
                context=context,
                lab=FecalOccultBloodTestFobt,
                title='Order a FOBT'))

        result.add_recommendation(
            LabRecommendation(
                key='CMS130v6_RECOMMEND_FITDNA',
                rank=2,
                button='Order',
                patient=self.patient,
                condition=MalignantNeoplasmOfColon,
                context=context,
                lab=FitDna,
                title='Order a FIT-DNA'))

        result.add_recommendation(
            ReferRecommendation(
                key='CMS130v6_RECOMMEND_SIGMOIDOSCOPY',
                rank=3,
                button='Order',
                patient=self.patient,
                referral=FlexibleSigmoidoscopy,
                context={
                    **context, 'specialties': ['Gastroenterology', ]
                },
                title='Order a Flexible sigmoidoscopy'))

        result.add_recommendation(
            ImagingRecommendation(
                key='CMS130v6_RECOMMEND_COLONOGRAPHY',
                rank=4,
                button='Order',
                patient=self.patient,
                imaging=CtColonography,
                context={
                    **context, 'specialties': ['Radiology', ]
                },
                title='Order a CT Colonography'))

        result.add_recommendation(
            ReferRecommendation(
                key='CMS130v6_RECOMMEND_COLONOSCOPY',
                rank=5,
                button='Order',
                patient=self.patient,
                referral=Colonoscopy,
                context={
                    **context, 'specialties': ['Gastroenterology', ]
                },
                title='Order a Colonoscopy'))

        return result



from typing import Any, Dict

def ensure_patient_in_group(patient_key: str, group_externally_exposable_id: str) -> Dict[str, Any]:
    
    return {
        'integration_type': 'Group',
        'integration_source': 'SDK',
        'handling_options': {
            'respond_async': True
        },
        'integration_payload': {
            'externally_exposable_id': group_externally_exposable_id,
            'group_type': 'person',
            'members': {
                'mode': 'appendOrRemove',
                'entries': [{
                    'uuid': patient_key,
                }]
            }
        }
    }

def ensure_patient_not_in_group(patient_key: str, group_externally_exposable_id: str) -> str:
    return {
        'integration_type': 'Group',
        'integration_source': 'SDK',
        'handling_options': {
            'respond_async': True
        },
        'integration_payload': {
            'externally_exposable_id': group_externally_exposable_id,
            'group_type': 'person',
            'members': {
                'mode': 'appendOrRemove',
                'entries': [{
                    'uuid': patient_key,
                    'inactive': True
                }]
            }
        }
    }
    